#------------------------------------------------------------------
# $Copyright: Copyright (c) 2014 Symantec Corporation.
# All rights reserved.
#
# THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF
# SYMANTEC CORPORATION.  USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED
# WITHOUT THE PRIOR EXPRESS WRITTEN PERMISSION OF SYMANTEC CORPORATION.
#
# The Licensed Software and Documentation are deemed to be commercial
# computer software as defined in FAR 12.212 and subject to restricted
# rights as defined in FAR Section 52.227-19 "Commercial Computer
# Software - Restricted Rights" and DFARS 227.7202, "Rights in
# Commercial Computer Software or Commercial Computer Software
# Documentation", as applicable, and any successor regulations. Any use,
# modification, reproduction release, performance, display or disclosure
# of the Licensed Software and Documentation by the U.S. Government
# shall be solely in accordance with the terms of this Agreement.  $
#
# $Header: /project/isv-agent/VProAgents/Tuxedo/Tuxedo.pm,v 1.22 2011/08/05 12:42:25 sgadgil Exp $
#
# $Author: sgadgil $
# $Date: 2011/08/05 12:42:25 $
# $Revision: 1.22 $
#
# Provides:                            Tuxedo
# Author:                              Anshuman Atre
# CrAgent Run Date:                    12 Jun, 2007
# Modification History:
# Known bugs/caveats:
#------------------------------------------------------------------

#------------
# Be strict..
#------------
use strict;

#----------------------
# Declare the package..
#----------------------
package Tuxedo;

#--------------------------------
# Include standard PERL modules..
#--------------------------------
use Fcntl;
use Socket;
use POSIX qw ( uname );

use IO::File;
use IO::Handle;

use File::stat;
use File::Path qw ( mkpath );

#-----------------------------------------
# Define module-vendor specific commands..
#-----------------------------------------
use ag_i18n_inc;

#-------------------------------------
# Include required framework modules..
#-------------------------------------
use Sys;            # Support generalized UNIX commands
use Proc;           # For calls to the process routines
use Comms;          # For calls to socket(3C) routines
use Version;        # Version display generator
use VCSagentFW;     # Primary Agent Interface to the ACC Library
use VCSDefines;     # This contains constants that VCS uses

my ( $CLUS_HOME, $sPrefix ) = ();

#---------------------
# Setup VCS/VAD home..
#---------------------
if ( $ENV { CLUSTER_HOME } ) { # VCS 5.0/VAD..
  $CLUS_HOME ||= $ENV { CLUSTER_HOME } ||= "/opt/VRTSvad";
}
else { # VCS
  $CLUS_HOME ||= $ENV { VCS_HOME } ||= "/opt/VRTSvcs";
}

( $sPrefix ) = $CLUS_HOME =~ /(...)$/;

my $sCmdHares  = $HARES ||= "$CLUS_HOME/bin/hares";
my $sCmdHatype = $HATYPE ||= "$CLUS_HOME/bin/hatype";
my $sCmdVcsDecrypt = $HADECRYPT ||= "$CLUS_HOME/sbin/${sPrefix}decrypt";

#-------------------------
# Define rsh/ssh command..
#-------------------------
my $sRshCommand = undef;
use constant RSH_COMMAND => \$sRshCommand;

my $bSharedLogin = 0; # Default: Unique..
my @liRunningProcs = (); # Global..

#-----------------------------------------------------------------------
# Add our version number to the version hash and check if versions are
# requested..
#-----------------------------------------------------------------------
my $sVer = "\$Id: Tuxedo.pm,v 1.22 2011/08/05 12:42:25 sgadgil Exp $::ETVHeader";
Version::TellVersion ( $sVer );

#--------------------------------
# Define Tuxedo subroutines..
#--------------------------------

=pod

 -------------------------------------------------------------------------
 Function :  bool ArgsValid ( IN String EntryPointName,
                              IN String EnvFile,
                              IN String MachineName,
                              IN String MachineType,
                              IN String RemoteShell,
                              IN String TmbootOptMaster,
                              IN String TmbootOptNonMaster,
                              IN String TmshutdownOptMaster,
                              IN String TmshutdownOptNonMaster,
                              IN String AppDir,
                              IN String TuxDir,
                              IN String TuxDomain,
                              IN String TuxServers,
                              IN String TuxServices,
                              IN String User,
                              IN Integer SecondLevelMonitor,
                              IN String MonitorProgram );

 Description : Validates the ArgList passed to the entry point.

 Parameters :

 $sEntryPointName               Entry pt name
 $localEnvFile                  Absolute path to environment file to source
 $localMachineName              Logical Machine ID of this instance
 $localMachineType              Bulletin board architecture
 $localRemoteShell              Absolute path to rsh/ssh binary
 $localTmbootOptMaster          Options for starting Master/Backup machine
 $localTmbootOptNonMaster       Options for starting NonMaster/Single machine
 $localTmshutdownOptMaster      Options for shutting down Master/Backup machine
 $localTmshutdownOptNonMaster   Options for shutting down NonMaster/Single machine
 $localAppDir                   Absolute path to the Application directory
 $localTuxDir                   Absolute path to Tuxedo base installation directory
 $localTuxDomain                Name of the Tuxedo domain for this instance
 $localTuxServers               Names of the Tuxedo servers to monitor
 $localTuxServices              Names of the Tuxedo services to monitor
 $localUser                     The dedicated UNIX login for this Tuxedo instance
 $localSecondLevelMonitor       Cycle count to run in-depth monitoring
 $localMonitorProgram           Full path to external monitor program
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub ArgsValid {
  #----------------------------------------
  # Set up local vars from passed in ones..
  #----------------------------------------
  my ( $sEntryPointName,
       $localEnvFile,
       $localMachineName,
       $localMachineType,
       $localRemoteShell,
       $localTmbootOptMaster,
       $localTmbootOptNonMaster,
       $localTmshutdownOptMaster,
       $localTmshutdownOptNonMaster,
       $localAppDir,
       $localTuxDir,
       $localTuxDomain,
       $localTuxServers,
       $localTuxServices,
       $localUser,
       $localSecondLevelMonitor,
       $localMonitorProgram ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20001, "Subroutine <%s> called with:
  EntryPointName           [%s]
  EnvFile                  [%s]
  MachineName              [%s]
  MachineType              [%s]
  RemoteShell              [%s]
  TmbootOptMaster          [%s]
  TmbootOptNonMaster       [%s]
  TmshutdownOptMaster      [%s]
  TmshutdownOptNonMaster   [%s]
  AppDir                   [%s]
  TuxDir                   [%s]
  TuxDomain                [%s]
  TuxServers               [%s]
  TuxServices              [%s]
  User                     [%s]
  SecondLevelMonitor       [%s]
  MonitorProgram           [%s]",

  $sMyName,  $sEntryPointName,  $localEnvFile,   $localMachineName,
  $localMachineType,    $localRemoteShell,   $localTmbootOptMaster,
  $localTmbootOptNonMaster,              $localTmshutdownOptMaster,
  $localTmshutdownOptNonMaster,     $localAppDir,     $localTuxDir,
  $localTuxDomain, $localTuxServers, $localTuxServices, $localUser,
  $localSecondLevelMonitor, $localMonitorProgram );

  #-------------------------------------
  # Set return code to hopeful success..
  #-------------------------------------
  my $iReturnCode = 1;

  #------------------------------------------
  # Get relevant references out of arg list..
  #------------------------------------------
  my $rsTmbootOptMaster = \$_[5]; # Holds localTmbootOptMaster..
  my $rsTuxServers = \$_[12];     # Holds localTuxServers..
  my $rsUser = \$_[14];           # Holds localUser..

  #--------------------------------------------
  # Create hash to weed out empty args values..
  #--------------------------------------------
  my ( @lsMissing, %hsArgList ) = ();
  my @lsArgs =
  qw ( EntryPointName  EnvFile   MachineName   MachineType   RemoteShell
       TmbootOptMaster       TmbootOptNonMaster      TmshutdownOptMaster
       TmshutdownOptNonMaster   AppDir   TuxDir   TuxDomain   TuxServers
       TuxServices User SecondLevelMonitor MonitorProgram );

  @hsArgList { @lsArgs } = @_; # Assign..

  #----------------------------------------
  # Reqd. attributes based on entry point..
  #----------------------------------------
  my @lsOnlineArgs  = qw ( AppDir EnvFile MachineName MachineType TmbootOptNonMaster
                           TmshutdownOptNonMaster TuxDir TuxDomain User );
  my @lsOfflineArgs = qw ( AppDir EnvFile MachineName MachineType TmshutdownOptNonMaster
                           TuxDir TuxDomain User );
  my @lsMonitorArgs = qw ( AppDir MachineName User );
  my @lsCleanArgs   = qw ( AppDir MachineName User );

  my %hlsNeeded = ( 'online'  => \@lsOnlineArgs,
                    'offline' => \@lsOfflineArgs,
                    'monitor' => \@lsMonitorArgs,
                    'clean'   => \@lsCleanArgs );

  my @lsReqdArgs = @{ $hlsNeeded { $sEntryPointName } };

  #-----------------------
  # Check what's missing..
  #-----------------------
  if ( @lsMissing = grep { $hsArgList { $_ } !~ /\S/o } @lsReqdArgs ) {
    &logError ( 20002,
    "<%s> The following mandatory fields were not set [%s], Total [%s]",
    $sMyName, join ( ' ', @lsMissing ), scalar @lsMissing );
    $iReturnCode = 0; # Failure..
  }

  $iReturnCode || return $iReturnCode; # Return, if failed..

  #-----------------------------
  # Remove SLM file, if exists..
  #-----------------------------
  if ( $sEntryPointName ne 'monitor' ) {
    VCSagentFW::RunSecondLevelMonitor ( 1 ); # Delete..
  }

  #-------------------------------
  # Individual validation starts..
  #-------------------------------

  #---------------------------
  # Check if login is shared..
  #---------------------------
  ++ $bSharedLogin if ( $localUser =~ s/^\+// ); # Set global..

  #------------------------
  # Validate parameter User..
  #------------------------
  if ( ! defined getpwnam $localUser ) {
    &logError ( 20003, "<%s> User [%s] is not a valid UNIX login",
    $sMyName, $localUser );
    $iReturnCode = 0; # Failure..
  }
  elsif ( $bSharedLogin ) {
    &logTrace ( 20004,
    "<%s> User [%s] is not unique to this VCS resource",
    $sMyName, $localUser );
    $$rsUser = $localUser; # Update..
  }


  #------------------
  # Return if clean..
  #------------------
  return $iReturnCode if ( $sEntryPointName eq 'clean' );

  #--------------------------------------------------------
  # Remove BBL and DBBL from list of servers if specified..
  #--------------------------------------------------------
  if ( $localTuxServers =~ /\b(D{0,1})BBL\b/ ){
    $$rsTuxServers = join ( ',', grep { ! /\b(^D{0,1})BBL\b/ }
                                 $localTuxServers =~ /([^(,|\s)]+)/g );
    &logWarn ( 20005,
    "<%s> Found BBL/DBBL in TuxServers [%s]. Changing to [%s]",
    $sMyName, $localTuxServers, $$rsTuxServers );
  }

  #--------------------
  # Return if monitor..
  #--------------------
  return $iReturnCode if ( $sEntryPointName eq 'monitor' );

  #------------------------------
  # Both online/offline allowed..
  #------------------------------
  #-----------------------------
  # Validate parameter EnvFile..
  #-----------------------------
  if ( ! -e  $localEnvFile ) {
    &logError ( 20006, "<%s> EnvFile [%s] file does not exist",
    $sMyName, $localEnvFile );
    $iReturnCode = 0; # Failure..
  }

  #----------------------------
  # Validate parameter AppDir..
  #----------------------------
  if ( ! -d  $localAppDir ) {
    &logError ( 20007, "<%s> AppDir [%s] directory does not exist",
    $sMyName, $localAppDir );
    $iReturnCode = 0; # Failure..
  }

  #----------------------------
  # Validate parameter TuxDir..
  #----------------------------
  if ( ! -d  $localTuxDir ) {
    &logError ( 20008, "<%s> TuxDir [%s] directory does not exist",
    $sMyName, $localTuxDir );
    $iReturnCode = 0; # Failure..
  }

  #-----------------------
  # Validate MachineType..
  #-----------------------
  if ( $localMachineType !~ /^((Non|)Master|Single|Backup)$/o ) {
    &logError ( 20009,
    "<%s> MachineType [%s] must be one of the following:
    Single, Master, NonMaster or Backup.",
    $sMyName, $localMachineType );
    $iReturnCode = 0; # Failure..
  }
  elsif ( ! Tuxedo::ValidateMachineTypeInDomain ( $localMachineType,
                                                  $localTuxDomain ) ) {
    &logError ( 20010,
    "<%s> Either multiple or no resources defined in TuxDomain [%s]
    as MachineType Master or Backup", $sMyName, $localTuxDomain );
  }

  #----------------------------------------
  # Verify rsh(1)/ssh(1) has been defined..
  #----------------------------------------
  if ( $localMachineType ne 'Single' && ! -x $localRemoteShell ) {
    &logError ( 20011,
    "<%s> If MachineType [%s] is not Single, RemoteShell [%s] has to be executable",
    $sMyName, $localMachineType, $localRemoteShell );
    $iReturnCode = 0; # Failure..
  }
  else {
    $sRshCommand = $localRemoteShell; # Populate RSH_COMMAND..
  }

  #------------------------------------
  # Validation for online entry point..
  #------------------------------------
  if ( $sEntryPointName eq "online" ) {
    #-----------------------------------------------------------
    # Do not allow -b option to be included in TmbootOptMaster..
    #-----------------------------------------------------------
    if ( grep { /^-/ && /b/ } $localTmbootOptMaster =~ /([\S]+)/g ) {
      $$rsTmbootOptMaster =
      join ( ' ', grep {
                    ( s/^-([^b]?)b+([^b]?)/-$1$2/g || 1 ) && # Exclude the bees..
                      ! /^-$/                                # Exclude a sole hyphen..
                  } $localTmbootOptMaster =~ /([\S]+)/g      # From what's been sent..
      );

      &logWarn ( 20012,
      "<%s> Option -b cannot be included in TmbootOptMaster [%s].
      Changing to [%s]",
      $sMyName, $localTmbootOptMaster, $$rsTmbootOptMaster );
    }

    #------------------------------------
    # Validate parameter MonitorProgram..
    #------------------------------------
    if ( $localMonitorProgram =~ /(\S+)/ && # Defined..
         ! -x $1 ) {
      &logError ( 20013, "<%s> MonitorProgram [%s] is not executable",
      $sMyName, $1 );
      $iReturnCode = 0; # Failure..
    }
  }

  $iReturnCode ||
  &logWarn ( 20014,
  "<%s> Some of the attributes could not be validated. Returning [%s]",
  $sMyName, $iReturnCode );

  return $iReturnCode ;
}



=pod

 -------------------------------------------------------------------------
 Function : bool Start ( IN String AppDir,
                         IN String EnvFile,
                         IN String MachineName,
                         IN String MachineType,
                         IN String TmbootOptMaster,
                         IN String TmbootOptNonMaster,
                         IN String TuxDir,
                         IN String TuxDomain,
                         IN String TuxServers,
                         IN String User
                         IN String ResourceName );

 Description : Start logic needed for online entry pt

 Parameters :

 $sAppDir                   Absolute path to the Application directory
 $sEnvFile                  Absolute path to environment file to source
 $sMachineName              Logical Machine ID of this instance
 $sMachineType              Bulletin board architecture
 $sTmbootOptMaster          Options for starting Master/Backup machine
 $sTmbootOptNonMaster       Options for starting NonMaster/Single machine
 $sTuxDir                   Absolute path to Tuxedo base installation directory
 $sTuxDomain                Name of the Tuxedo domain for this instance
 $sTuxServers               Names of the Tuxedo servers to monitor
 $sUser                     The dedicated UNIX login for this Tuxedo instance
 $sResourceName             Name of this resource
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub Start {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sAppDir,
       $sEnvFile,
       $sMachineName,
       $sMachineType,
       $sTmbootOptMaster,
       $sTmbootOptNonMaster,
       $sTuxDir,
       $sTuxDomain,
       $sTuxServers,
       $sUser,
       $sResourceName ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20015, "Subroutine <%s> called with:
  AppDir                   [%s]
  EnvFile                  [%s]
  MachineName              [%s]
  MachineType              [%s]
  TmbootOptMaster          [%s]
  TmbootOptNonMaster       [%s]
  TuxDir                   [%s]
  TuxDomain                [%s]
  TuxServers               [%s]
  User                     [%s]
  ResourceName             [%s]",

  $sMyName,  $sAppDir,  $sEnvFile,  $sMachineName,   $sMachineType,
  $sTmbootOptMaster,  $sTmbootOptNonMaster,  $sTuxDir, $sTuxDomain,
  $sTuxServers, $sUser, $sResourceName );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;
  my $sNodeName = ( uname )[1]; # Grab second element..

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sCommand, $sCommandArgs,
       $iTillWhen, $iExitCode, $iTimeout, $iMstrSrchLmt,
       @lsCmdout, @liPids ) = ();

  #-----------------------------------------
  # Whack any preexisting Tuxedo processes..
  #-----------------------------------------
  &logTrace ( 20016,
  "<%s> Killing Tuxedo processes for this instance", $sMyName );

  if ( Tuxedo::KillTuxedoProcesses ( $sAppDir,
                                     $sMachineName,
                                     $sTuxServers,
                                     $sUser ) ) {
    &logTrace ( 20017,
    "<%s> Processes for this instance do not exist before start-up",
    $sMyName );
  }
  else {
    &logError ( 20018,
    "<%s> Could not remove all processes for this instance. Exiting",
    $sMyName );
    $iReturnCode = 0; # Failure..
  }

  #---------------------------------------------------------------------
  # Upper bound.. Time in EPOCH when entry pt started + VCS time limit..
  #---------------------------------------------------------------------
  $iTillWhen = $^T + Tuxedo::GetEngineTimeout ( 1 );
  $iTimeout = $iTillWhen - time;
  
  #-----------------------------------
  # Remove preexisting IPC resources..
  #-----------------------------------
  if ( 1 == ( $iExitCode = Tuxedo::RemoveIPCs ( $sEnvFile,
                                                $sTuxDir,
                                                $sUser,
                                                $iTimeout ) ) ) {
    &logTrace ( 20019,
    "<%s> Removed IPC resources for Tuxedo machine [%s]",
    $sMyName, $sMachineName );

    &logTrace ( 20020,
    "<%s> Removed IPC resources for Tuxedo machine [%s] before startup",
    $sMyName, $sMachineName );
  }
  elsif ( -1 == $iExitCode ) { # Critical error..
    &logError ( 20021,
    "<%s> IPC resources could not be cleared for Tuxedo machine [%s]. Exiting.",
    $sMyName, $sMachineName );

    $iReturnCode = 0; # Failure..
  }
  else {
    &logTrace ( 20022,
    "<%s> IPC resources for Tuxedo machine [%s] do not exist before startup",
    $sMyName, $sMachineName );
  }

  #--------------------
  # Return, if failed..
  #--------------------
  $iReturnCode || return $iReturnCode;

  ( @lsCmdout, $iExitCode, $iTimeout ) = (); # Ensure..

  #---------------------------------------------
  # Determine whether we are in SHM or MP mode..
  #---------------------------------------------
  if ( $sMachineType eq 'Single' ) { # SHM..
    #------------------------------------------
    # Define start command and args variables..
    #------------------------------------------
    $sCommand = "$sTuxDir/bin/tmboot";
    $sCommandArgs = "$sTmbootOptNonMaster";
  
    &logTrace ( 20023,
    "<%s> Starting Tuxedo instance with Cmd [%s] Args [%s]",
    $sMyName, $sCommand, $sCommandArgs );
  
    #------------------
    # Request startup..
    #------------------
    if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
      &logError ( 20024,
      "<%s> Insufficient time to continue boot up. Short by [%s] sec",
      $sMyName, abs ( $iTimeout ) );
  
      $iReturnCode = 0; # Failure..
    }
    elsif ( $iReturnCode =
      Tuxedo::ExecLocalCmd ( $sUser, 
                             $sEnvFile, 
                             $sCommand, 
                             $sCommandArgs, 
                             \$iExitCode, 
                             $iTimeout, 
                             \@lsCmdout ) ) {
      &logTrace ( 20025,
      "<%s> Cmd [%s] with args [%s] executed successfully",
      $sMyName, $sCommand, $sCommandArgs );
    }
  }
  else { # MP..
    #--------------------------------------------------
    # Attributes of interest for the active tuxmaster..
    #--------------------------------------------------
    my ( $sMstrRsrce, $sMstrEnvFile, $sMstrMcName, $sMstrMcType,
         $sMstrTxDr, $sMstrTxDmn, $sMstrUsr, $sMstrSysName,
         @lsMstrAttr ) = ();

    #---------------------------------------
    # Find the active master in the domain..
    #---------------------------------------
    my $iAttempts = 0;
    my $iRetryPause = 5; # Seconds..
    $iMstrSrchLmt = Tuxedo::GetMasterSearchWaitLimit ( $sMachineType,
                                                       $iTillWhen );

    while ( ( $iTimeout = $iMstrSrchLmt - time ) > 1 ) {
      ++ $iAttempts;         # Increment attempt counter..

      #-------------------------------
      # Search for the active master..
      #-------------------------------
      &logTrace ( 20026,
      "<%s> +----- Searching for active master with [%s] sec. to spare -----+",
      $sMyName, $iTimeout );

      if ( Tuxedo::WhereIsMasterActive ( $sTuxDomain,
                                         $sTuxDir,
                                         \@lsMstrAttr ) ) {
        #--------------------------------------------------------
        # Get required attributes of the active master resource..
        #--------------------------------------------------------
        ( $sMstrSysName,
          $sMstrEnvFile,
          $sMstrMcName,
          $sMstrMcType,
          $sMstrTxDr,
          $sMstrTxDmn,
          $sMstrUsr,
          $sMstrRsrce ) = @lsMstrAttr;

        &logTrace ( 20027, "<%s> Found active master resource [%s] with:
        EnvFile     [%s]
        MachineName [%s]
        MachineType [%s]
        TuxDir      [%s]
        TuxDomain   [%s]
        User        [%s] online on
        VCS Node    [%s]",
        $sMyName, $sMstrRsrce, $sMstrEnvFile, $sMstrMcName, $sMstrMcType,
        $sMstrTxDmn, $sMstrTxDr, $sMstrUsr, $sMstrSysName );

        $iAttempts = 0; # Reset to indicate success..
        last;           # Found.. get out..
      }
      else {
        &logTrace ( 20028,
        "<%s> Failed to find active Master in domain [%s] in attempt [%s]",
        $sMyName, $sTuxDomain, $iAttempts );

        sleep $iRetryPause; # Pause before attempting again..
      }
    }

    #-------------------------------
    # Failed to find active master..
    #-------------------------------
    if ( $iAttempts ) {
      if ( $sMachineType =~ /^(Master|Backup)$/o ) {
        &logInfo ( 20029,
        "<%s> Could not locate active Master for domain [%s] after [%s] attempts.
         Booting current resource as Active Master",
         $sMyName, $sTuxDomain, $iAttempts );

        my $iBootCnt = 0;
        my $sBootStr = Tuxedo::UseBackupBootFlag ( $sUser,
                                                   $sEnvFile,
                                                   $sMachineName )?'-b ':'';
        while ( $iBootCnt <= 1 ) {
          if ( $iBootCnt ) {
            $sBootStr = $sBootStr eq '-b '?'':'-b '; 
            &logInfo ( 20030,
            "<%s> Previous attempt to boot the active master failed.
            Making another attempt with boot flag [%s]", $sMyName, $sBootStr );
          }

          #----------------------------------------------------------
          # Boot resource as the new active master.. i.e. host DBBL..
          #----------------------------------------------------------
          $sCommand = "$sTuxDir/bin/tmboot";
          $sCommandArgs = "$sBootStr$sTmbootOptMaster";

          #--------------------
          # Get time bearings..
          #--------------------
          if ( ( $iTimeout = $iTillWhen - time ) <= 1 ) {
            &logError ( 20031,
            "<%s> Insufficient time to continue boot up. Short by [%s] sec",
            $sMyName, abs ( $iTimeout ) );
            return 0; # Failure..
          }
  
          &logTrace ( 20032,
          "<%s> Starting Tuxedo services with Command [%s] Args [%s]",
          $sMyName, $sCommand, $sCommandArgs );

          #------------------------
          # Request local startup..
          #------------------------
          $iReturnCode = Tuxedo::ExecLocalCmd ( $sUser,
                                                $sEnvFile,
                                                $sCommand,
                                                $sCommandArgs,
                                                \$iExitCode,
                                                $iTimeout,
                                                \@lsCmdout );

          #------------------------------------------------------------------
          # At this time, if we successfully managed to execute the tmboot(1)
          # command,    we    need   to   update   the   history   file   for
          # WhereIsMasterActive. A monitor is  immediately  going  to  follow
          # this  return.   An  incorrect  history  file  is  going  to force
          # WhereIsMasterActive to use  the  hares(1m)  option  to  locate  a
          # system  to  run its psr based DBBL search. Since we aren't online
          # yet, it would fail. This would *force* a migration even though we
          # have  the  DBBL.  A  migration  command  in NOT available for the
          # active master. An update would make it run the psr check locally.
          #------------------------------------------------------------------

          if ( $iReturnCode ) {
            #----------------------------------------------
            # Update history file for WhereIsMasterActive..
            #----------------------------------------------
            @lsMstrAttr = ( $sNodeName, $sEnvFile, $sMachineName,
                            $sMachineType, $sTuxDir, $sTuxDomain, $sUser,
                            $sResourceName );
  
            $lsMstrAttr [-2] = "+$sUser" if ( $bSharedLogin ); # Shared login..

            if ( Tuxedo::UpdateTuxMasterHistory ( $sTuxDir,
                                                  \@lsMstrAttr ) ) {
              &logTrace ( 20033, "<%s> Updated history file successfully",
              $sMyName );
            }
            else {
              &logError ( 20034, "<%s> Failed to update history file", $sMyName );
            }

            last; # Break loop, if successful..
          } 

          ++ $iBootCnt;
        }
      }
      else { # MachineType is NonMaster..
        &logError ( 20035,
        "<%s> Could not locate active Master for domain [%s] after [%s] attempts.
        This NonMaster instance will not be started",
        $sMyName, $sTuxDomain, $iAttempts );

        $iReturnCode = 0; # Failure..
      }
    }
    else { # We found the active master..
      #------------------------------------------
      # First request a bulletin board clean up..
      #------------------------------------------
      @lsCmdout = (); # Ensure..

      #--------------------
      # Get time bearings..
      #--------------------
      if ( ( $iTimeout = $iTillWhen - time ) <= 1 ) {
        &logError ( 20036,
        "<%s> Insufficient time to continue boot up. Short by [%s] sec",
        $sMyName, abs ( $iTimeout ) );
        return 0; # Failure..
      }
      
      &logTrace ( 20037, "<%s> Cleaning Tuxedo services for Tuxedo machine [%s]",
      $sMyName, $sMachineName );

      $iReturnCode = Tuxedo::CleanBBofMachine ( $sMstrEnvFile,
                                                $sMstrTxDr,
                                                $sMachineName,
                                                $sMstrUsr,
                                                \@lsCmdout,
                                                $iTimeout,
                                                $sMstrSysName );

      #------------------------------------------------
      # Handle return code.. mostly to ignore failure..
      #------------------------------------------------
      if ( $iReturnCode ) { # Success..
        &logTrace ( 20038, "<%s> Cleaned up bulletin board successfully.",
        $sMyName );
      }
      else { # Failure.. check the reason..
        #---------------------------------------------------------------------
        # We had requested a BBL clean up as a precaution. If the clean up was
        # not at all needed, we'd fail in a successful execution, and hence we
        # can neglect the failed exit code.
        #---------------------------------------------------------------------
        $iReturnCode = 1; # Success..
      }

      $iReturnCode || return $iReturnCode; # Return if failed..

      #----------------------------------------------
      # Now request the active master for a boot up..
      #----------------------------------------------
      $sCommand = "$sMstrTxDr/bin/tmboot"; # Reset..
      $sCommandArgs = $sTmbootOptNonMaster;  # Boot up as a non-master..

      #--------------------
      # Get time bearings..
      #--------------------
      if ( ( $iTimeout = $iTillWhen - time ) <= 1 ) {
        &logError ( 20039,
        "<%s> Insufficient time to continue boot up. Short by [%s] sec",
        $sMyName, abs ( $iTimeout ) );
        return 0; # Failure..
      }

      &logTrace ( 20040,
      "<%s> Starting Tuxedo services with Command [%s] Args [%s]",
      $sMyName, $sCommand, $sCommandArgs );

      #----------------------------------
      # Now fire command on remote node..
      #----------------------------------
      $iReturnCode = Tuxedo::ExecRemoteTuxCmd ( $sMstrEnvFile,
                                                $sMstrUsr,
                                                $sCommand,
                                                $sCommandArgs,
                                                \@lsCmdout,
                                                $iTimeout,
                                                $sMstrSysName );
    }
  }

  #-------------------
  # Return if failed..
  #-------------------
  $iReturnCode || return $iReturnCode;

  #--------------------
  # Delay entry point..
  #--------------------
  my $iSleepTime = 10; # Secs..
  
  while ( ( $iTimeout = $iTillWhen - time ) > 10 ) {
    @liPids = (); # Initialise..

    sleep $iSleepTime; # Sec..

    &logTrace ( 20041,
    "<%s> Confirming start-up with [%s] sec in hand",
    $sMyName, $iTimeout );

    #---------
    # Do FLM..
    #---------
    if ( VCSDef_MonitorExit::ONLINE ==
         Tuxedo::GetProcState ( \@liPids,
                                $sAppDir,
                                $sMachineName,
                                $sTuxServers,
                                $sUser,
                                1 ) ) { # Fast..
      #---------------------
      # Start-up confirmed..
      #---------------------
      &logInfo ( 20042,
      "<%s> Confirmed start-up of this instance of Tuxedo server",
      $sMyName );

      $iReturnCode = 1; # Success..
      last;
    }
    else {
      &logTrace ( 20043,
      "<%s> Could not confirm start-up. Sleeping for [%s] sec",
      $sMyName, $iSleepTime );

      $iReturnCode = 0;  # Failure..
    }
  }

  #---------
  # Re-set..
  #---------
  $iTimeout = $iTillWhen - time;
  @lsCmdout = ();

  #--------------------------------------
  # Create IPC history, if login shared..
  #--------------------------------------
  if ( $bSharedLogin &&
       Tuxedo::ManageTuxIPCs ( $sEnvFile,
                               $sTuxDir,
                               $sUser,
                               $iTimeout,
                               0,            # Create history..
                               \@lsCmdout ) ) {
    &logTrace ( 20044,
    "<%s> Created IPC history file successfully",
    $sMyName );
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool Stop ( IN String AppDir,
                        IN String EnvFile,
                        IN String MachineName,
                        IN String MachineType,
                        IN String TmshutdownOptMaster,
                        IN String TmshutdownOptNonMaster,
                        IN String TuxDir,
                        IN String TuxDomain,
                        IN String TuxServers,
                        IN String User );

 Description : Stop logic needed for offline entry pt

 Parameters :

 $sAppDir                   Absolute path to the Application directory
 $sEnvFile                  Absolute path to environment file to source
 $sMachineName              Logical Machine ID of this instance
 $sMachineType              Bulletin board architecture
 $sTmshutdownOptMaster      Options for shutting down Master/Backup machine
 $sTmshutdownOptNonMaster   Options for shutting down NonMaster/Single machine
 $sTuxDir                   Absolute path to Tuxedo base installation directory
 $sTuxDomain                Name of the Tuxedo domain for this instance
 $sTuxServers               Names of the Tuxedo servers to monitor
 $sUser                     The dedicated UNIX login for this Tuxedo instance
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub Stop {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sAppDir,
       $sEnvFile,
       $sMachineName,
       $sMachineType,
       $sTmshutdownOptMaster,
       $sTmshutdownOptNonMaster,
       $sTuxDir,
       $sTuxDomain,
       $sTuxServers,
       $sUser ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20045, "Subroutine <%s> called with:
  AppDir                   [%s]
  EnvFile                  [%s]
  MachineName              [%s]
  MachineType              [%s]
  TmshutdownOptMaster      [%s]
  TmshutdownOptNonMaster   [%s]
  TuxDir                   [%s]
  TuxDomain                [%s]
  TuxServers               [%s]
  User                     [%s]",

  $sMyName,  $sAppDir,  $sEnvFile,  $sMachineName,   $sMachineType,
  $sTmshutdownOptMaster,     $sTmshutdownOptNonMaster,    $sTuxDir,
  $sTuxDomain, $sTuxServers, $sUser );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1; # Success..

  #---------------------------------------------------------------------
  # Upper bound.. Time in EPOCH when entry pt started + VCS time limit..
  #---------------------------------------------------------------------
  my $iTillWhen = $^T + Tuxedo::GetEngineTimeout ( 1 ); # With margin..

  #------------------------------------------
  # Declare stop command and args variables..
  #------------------------------------------
  my ( $sCommand, $sCommandArgs, $iTimeout, $iExitCode,
       $iSleepTime, $bFlag, @lsCmdout, @liPids ) = ();
  
  #--------------------------------------------------
  # Attributes of interest for the active tuxmaster..
  #--------------------------------------------------
  my ( $iMstrSrchLmt, $sMstrRsrce, $sMstrEnvFile, $sMstrMcName,
       $sMstrMcType, $sMstrTxDr, $sMstrTxDmn, $sMstrUsr,
       $sMstrSysName, $sBkupType, @lsMstrAttr ) = ();

  #---------------------------------------------
  # Determine whether we are in SHM or MP mode..
  #---------------------------------------------
  if ( $sMachineType eq 'Single' ) { # SHM..
    #------------------------------------
    # Define stop command and arguments..
    #------------------------------------
    $sCommand = "$sTuxDir/bin/tmshutdown";
    $sCommandArgs = "$sTmshutdownOptNonMaster";
  
    &logTrace ( 20046,
    "<%s> Stopping Tuxedo with Command [%s] Args [%s]",
    $sMyName, $sCommand, $sCommandArgs );
  
    #-------------------------
    # Request local shutdown..
    #-------------------------
    if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
      &logError ( 20047,
      "<%s> Insufficient time to continue shutdown. Short by [%s] sec",
      $sMyName, abs ( $iTimeout ) );
  
      $iReturnCode = 0; # Failure..
    }
    elsif ( $iReturnCode =
            Tuxedo::ExecLocalCmd ( $sUser, 
                                   $sEnvFile, 
                                   $sCommand, 
                                   $sCommandArgs, 
                                   \$iExitCode, 
                                   $iTimeout, 
                                   \@lsCmdout ) ) {
      &logTrace ( 20048,
      "<%s> Cmd [%s] with args [%s] executed successfully",
      $sMyName, $sCommand, $sCommandArgs );
    }
  }
  else { # MP..
    if ( Tuxedo::WhereIsMasterActive ( $sTuxDomain,
                                       $sTuxDir,
                                       \@lsMstrAttr ) ) {
      #--------------------------------------------------------
      # Get required attributes of the active master resource..
      #--------------------------------------------------------
      ( $sMstrSysName,
        $sMstrEnvFile,
        $sMstrMcName,
        $sMstrMcType,
        $sMstrTxDr,
        $sMstrTxDmn,
        $sMstrUsr,
        $sMstrRsrce ) = @lsMstrAttr;
  
      &logTrace ( 20049, "<%s> Found active master resource [%s] with:
      EnvFile     [%s]
      MachineName [%s]
      MachineType [%s]
      TuxDir      [%s]
      TuxDomain   [%s]
      User        [%s] online on
      VCS Node    [%s]",
      $sMyName,
      $sMstrRsrce, $sMstrEnvFile, $sMstrMcName, $sMstrMcType,
      $sMstrTxDr, $sMstrTxDmn, $sMstrUsr, $sMstrSysName );

      #--------------------
      # Get time bearings..
      #--------------------
      if ( ( $iTimeout = $iTillWhen - time ) <= 1 ) {
        &logError ( 20050, "<%s> Insufficient time to continue shutdown.
        Short by [%s] sec", $sMyName, abs ( $iTimeout ) );
        return 0; # Failure..
      }

      #------------------------------------------
      # Verify that we aren't the active master..
      #------------------------------------------
      if ( $sMstrMcName ne $sMachineName ) { # We are not..
        $sCommand = "$sMstrTxDr/bin/tmshutdown";
        $sCommandArgs = $sTmshutdownOptNonMaster;
 
        &logTrace ( 20051,
        "<%s> Stopping Tuxedo services with Command [%s] Args [%s]",
        $sMyName, $sCommand, $sCommandArgs );

        #----------------------------------
        # Now fire command on remote node..
        #----------------------------------
        $iReturnCode = Tuxedo::ExecRemoteTuxCmd ( $sMstrEnvFile,
                                                  $sMstrUsr,
                                                  $sCommand,
                                                  $sCommandArgs,
                                                  \@lsCmdout,
                                                  $iTimeout,
                                                  $sMstrSysName );
        ++ $bFlag; # Set flag..
      }
      else { # We are the active master..
        $sBkupType = $sMachineType eq 'Master' ? 'Backup' : 'Master';

        #------------------------------
        # Attempt to migrate the DBBL..
        #------------------------------
        if ( Tuxedo::MigrateDBBLToBkp ( $sBkupType,
                                        $sTuxDomain,
                                        $iTimeout,
                                        \@lsMstrAttr ) ) { # Success!

          #---------------------------------------------------
          # Get required attributes of the new active master..
          #---------------------------------------------------
          ( $sMstrSysName,
            $sMstrEnvFile,
            $sMstrMcName,
            $sMstrMcType,
            $sMstrTxDr,
            $sMstrTxDmn,
            $sMstrUsr,
            $sMstrRsrce ) = @lsMstrAttr;
      
          &logTrace ( 20052, "<%s> Found new active master resource [%s] with:
          EnvFile     [%s]
          MachineName [%s]
          MachineType [%s]
          TuxDir      [%s]
          TuxDomain   [%s]
          User        [%s] online on
          VCS Node    [%s]",
          $sMyName,
          $sMstrRsrce, $sMstrEnvFile, $sMstrMcName, $sMstrMcType,
          $sMstrTxDr, $sMstrTxDmn, $sMstrUsr, $sMstrSysName );

          $sCommand = "$sMstrTxDr/bin/tmshutdown";
          $sCommandArgs = $sTmshutdownOptNonMaster;
   
          &logTrace ( 20053,
          "<%s> Stopping Tuxedo services with Command [%s] Args [%s]",
          $sMyName, $sCommand, $sCommandArgs );
  
          #----------------------------------
          # Now fire command on remote node..
          #----------------------------------
          $iReturnCode = Tuxedo::ExecRemoteTuxCmd ( $sMstrEnvFile,
                                                    $sMstrUsr,
                                                    $sCommand,
                                                    $sCommandArgs,
                                                    \@lsCmdout,
                                                    $iTimeout,
                                                    $sMstrSysName );
          ++ $bFlag; # Set flag..
        }
        else { # Boo..
          #--------------------------------
          # Set up local shutdown options..
          #--------------------------------
          $sCommand = "$sTuxDir/bin/tmshutdown";
          $sCommandArgs = $sTmshutdownOptMaster; # Shutdown as active master..

          #--------------------
          # Get time bearings..
          #--------------------
          if ( ( $iTimeout = $iTillWhen - time ) <= 1 ) {
            &logError ( 20054, "<%s> Insufficient time to continue shutdown.
            Short by [%s] sec", $sMyName, abs ( $iTimeout ) );
            return 0; # Failure..
          }
   
          &logTrace ( 20055,
          "<%s> Stopping Tuxedo services with Command [%s] Args [%s]",
          $sMyName, $sCommand, $sCommandArgs );
  
          #-------------------------
          # Request local shutdown..
          #-------------------------
          Tuxedo::ExecLocalCmd ( $sUser, # Ignore error code for active master..
                                 $sEnvFile,
                                 $sCommand,
                                 $sCommandArgs,
                                 \$iExitCode,
                                 $iTimeout,
                                 \@lsCmdout );
        }
      }
    }
    else {
      #--------------------------------------------------------------
      # At this time, we don't have an active master to shut us down.
      # We are partitioned. Proceed accordingly..
      #--------------------------------------------------------------
      &logWarn ( 20056,
      "<%s> Failed to locate active master. Machine [%s] is PARTITIONED",
      $sMyName, $sMachineName );

      #--------------------------------
      # Set up local shutdown options..
      #--------------------------------
      $sCommand = "$sTuxDir/bin/tmshutdown";
      $sCommandArgs = "-P $sMachineName"; # Partioned..

      #--------------------
      # Get time bearings..
      #--------------------
      if ( ( $iTimeout = $iTillWhen - time ) <= 1 ) {
        &logError ( 20057, "<%s> Insufficient time to continue shutdown.
        Short by [%s] sec", $sMyName, abs ( $iTimeout ) );
        return 0; # Failure..
      }

      #-------------------------
      # Request local shutdown..
      #-------------------------
      $iReturnCode = Tuxedo::ExecLocalCmd ( $sUser,
                                            $sEnvFile,
                                            $sCommand,
                                            $sCommandArgs,
                                            \$iExitCode,
                                            $iTimeout,
                                            \@lsCmdout );
      ++ $bFlag; # Set flag..
    }
  }

  #----------------------------------------
  # Whack any persisting Tuxedo processes..
  #----------------------------------------
  &logTrace ( 20058,
  "<%s> Killing Tuxedo processes for this instance", $sMyName );

  if ( Tuxedo::KillTuxedoProcesses ( $sAppDir,
                                     $sMachineName,
                                     $sTuxServers,
                                     $sUser ) ) {
    &logTrace ( 20059,
    "<%s> Processes for this instance do not exist after shutdown",
    $sMyName );
  }
  else {
    &logError ( 20060,
    "<%s> Could not remove all processes for this instance. Exiting",
    $sMyName );

    $iReturnCode = 0; # Failure..
  }

  #----------------------------------
  # Remove persisting IPC resources..
  #----------------------------------
  if ( 1 == ( $iExitCode = Tuxedo::RemoveIPCs ( $sEnvFile,
                                                $sTuxDir,
                                                $sUser,
                                                $iTimeout ) ) ) {
    &logTrace ( 20061,
    "<%s> IPC resources for Tuxedo machine [%s] removed after shutdown",
    $sMyName, $sMachineName );
  }
  elsif ( -1 == $iExitCode ) {
    &logError ( 20062,
    "<%s> Failed to remove IPC resources for Tuxedo Machine [%s]", 
    $sMyName, $sMachineName );

    $iReturnCode = 0; # Failure..
  }
  else {
    &logTrace ( 20063,
    "<%s> IPC resources for Tuxedo Machine [%s] do not exist after shutdown",
    $sMyName, $sMachineName );
  }

  Tuxedo::DeletePidFile ( );

  return $iReturnCode if $sMachineType eq 'Single'; # SHM mode over..

  #-------------------------------------
  # Request clean up of bulletin board..
  #-------------------------------------
  ( $sMstrRsrce, $sMstrEnvFile, $sMstrMcName,        # Reset..
    $sMstrMcType, $sMstrTxDr, $sMstrTxDmn, $sMstrUsr,
    $sMstrSysName, @lsCmdout ) = ();

  my $iAttempts = 0;    # Search attempts..
  my $iRetryPause = 10; # Secs..

  #-----------------------------------------
  # Get search limit for the active master..
  #-----------------------------------------
  $iMstrSrchLmt = Tuxedo::GetMasterSearchWaitLimit ( $sMachineType,
                                                     $iTillWhen );

  while ( ( $iTimeout = $iMstrSrchLmt - time ) > 1 ) {
    &logTrace ( 20064,
    "<%s> +----- Searching for active master with [%s] sec. to spare -----+",
    $sMyName, $iTimeout );
  
    ++ $iAttempts; # Increment attempt counter..

    #-------------------------------
    # Search, if not found earlier..
    #-------------------------------
    if ( $bFlag || Tuxedo::WhereIsMasterActive ( $sTuxDomain,
                                                 $sTuxDir,
                                                 \@lsMstrAttr ) ) {
      $iAttempts = 0; # Found! Reset..
  
      #--------------------------------------------------------
      # Get required attributes of the active master resource..
      #--------------------------------------------------------
      ( $sMstrSysName,
        $sMstrEnvFile,
        $sMstrMcName,
        $sMstrMcType,
        $sMstrTxDr,
        $sMstrTxDmn,
        $sMstrUsr,
        $sMstrRsrce ) = @lsMstrAttr;
  
      &logTrace ( 20065,
      "<%s> Found active master resource [%s] with:
      EnvFile     [%s]
      MachineName [%s]
      MachineType [%s]
      TuxDir      [%s]
      TuxDomain   [%s]
      User        [%s] online on
      VCS Node    [%s]",
      $sMyName,
      $sMstrRsrce, $sMstrEnvFile, $sMstrMcName, $sMstrMcType,
      $sMstrTxDmn, $sMstrTxDr, $sMstrUsr, $sMstrSysName );
  
      #---------------------------------------------------
      # Request TuxMaster to clean up the bulletin board..
      #---------------------------------------------------
      if ( Tuxedo::CleanBBofMachine ( $sMstrEnvFile,
                                      $sMstrTxDr,
                                      $sMachineName,
                                      $sMstrUsr,
                                      \@lsCmdout,
                                      $iTimeout,
                                      $sMstrSysName ) ) {
        &logTrace ( 20066,
        "<%s> Cleaned up bulletin board successfully.",
        $sMyName );
      }

      last; # Get out..
    }
    else {
      &logTrace ( 20067,
      "<%s> Could not locate active master in domain [%s] for attempt [%s]",
      $sMyName, $sTuxDomain, $iAttempts );

      #--------------------------------
      # Pause before attempting again..
      #--------------------------------
      sleep $iRetryPause;
    }
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool Kill ( IN String AppDir,
                        IN Integer CleanReason,
                        IN String EnvFile,
                        IN String MachineName,
                        IN String MachineType,
                        IN String RemoteShell,
                        IN String TuxDir,
                        IN String TuxDomain,
                        IN String TuxServers,
                        IN String User );

 Description : Forces shutdown of resource needed for clean entry point

 Parameters :

 $sAppDir                   Absolute path to the Application directory
 $iCleanReason              Why clean was called
 $sEnvFile                  Absolute path to environment file to source
 $sMachineName              Logical Machine ID of this instance
 $sMachineType              Bulletin board architecture
 $sRemoteShell              Absolute path to rsh/ssh binary
 $sTuxDir                   Absolute path to Tuxedo base installation directory
 $sTuxDomain                Name of the Tuxedo domain for this instance
 $sTuxServers               Names of the Tuxedo servers to monitor
 $sUser                     The dedicated UNIX login for this Tuxedo instance
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub Kill {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sAppDir,
       $iCleanReason,
       $sEnvFile,
       $sMachineName,
       $sMachineType,
       $sRemoteShell,
       $sTuxDir,
       $sTuxDomain,
       $sTuxServers,
       $sUser ) = @_;

  #-----------------------------
  # Figure out why were called..
  #-----------------------------
  my ( $sCleanReason, %hsCleanReasons ) = ();
  my @lsCleanStrings = qw ( Offline_timed_out    Offline_was_ineffective
                            Online_timed_out      Online_was_ineffective
                            Found_offline_unexpectedly Monitor_timed_out );

  @hsCleanReasons { 0 .. $#lsCleanStrings } = @lsCleanStrings; # Populate..
  ( $sCleanReason = $hsCleanReasons { $iCleanReason } ||       # Reason..
                    $iCleanReason ) =~ s/_/ /g;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20068, "Subroutine <%s> called with:
  AppDir                   [%s]
  CleanReason              [%s]
  EnvFile                  [%s]
  MachineName              [%s]
  MachineType              [%s]
  RemoteShell              [%s]
  TuxDir                   [%s]
  TuxDomain                [%s]
  TuxServers               [%s]
  User                     [%s]",

  $sMyName,  $sAppDir,  $sCleanReason,  $sEnvFile,   $sMachineName,
  $sMachineType,      $sRemoteShell,     $sTuxDir,     $sTuxDomain,
  $sTuxServers, $sUser );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #---------------------------------------------------------------------
  # Upper bound.. Time in EPOCH when entry pt started + VCS time limit..
  #---------------------------------------------------------------------
  my $iTillWhen = $^T + Tuxedo::GetEngineTimeout ( 1 ); # With margin..

  #--------------------------
  # Declare local variables..
  #--------------------------
  my ( $iTimeout, $iExitCode, @lsCmdout ) = ();

  #--------------------------------------
  # Whack any Tuxedo processes directly..
  #--------------------------------------
  &logTrace ( 20069,
  "<%s> Killing Tuxedo processes for this instance", $sMyName );

  if ( Tuxedo::KillTuxedoProcesses ( $sAppDir,
                                     $sMachineName,
                                     $sTuxServers,
                                     $sUser ) ) {
    &logTrace ( 20070,
    "<%s> Processes for this instance do not exist after shutdown",
    $sMyName );
  }
  else {
    &logError ( 20071,
    "<%s> Could not remove all processes for this instance. Exiting",
    $sMyName );

    $iReturnCode = 0; # Failure..
  }

  #----------------------------------
  # Remove persisting IPC resources..
  #----------------------------------
  $iTimeout = $iTillWhen - time;

  #----------------------------------
  # Remove persisting IPC resources..
  #----------------------------------
  if ( 1 == ( $iExitCode = Tuxedo::RemoveIPCs ( $sEnvFile,
                                                $sTuxDir,
                                                $sUser,
                                                $iTimeout ) ) ) {
    &logTrace ( 20072,
    "<%s> IPC resources for Tuxedo machine [%s] removed after shutdown",
    $sMyName, $sMachineName );
  }
  elsif ( -1 == $iExitCode ) {
    &logError ( 20073,
    "<%s> Failed to remove IPC resources for Tuxedo Machine [%s]",
    $sMyName, $sMachineName );

    $iReturnCode = 0; # Failure..
  }
  else {
    &logTrace ( 20074,
    "<%s> IPC resources for Tuxedo Machine [%s] do not exist after shutdown",
    $sMyName, $sMachineName );
  }

  Tuxedo::DeletePidFile ( );

  return $iReturnCode if $sMachineType eq 'Single'; # SHM mode over..

  #----------------------------------------
  # Verify rsh(1)/ssh(1) has been defined..
  #----------------------------------------
  if ( -x $sRemoteShell ) {
    $sRshCommand = $sRemoteShell; # Populate RSH_COMMAND..
  }
  else {
    &logError ( 20075,
    "<%s> Remote shell binary [%s] not executable.",
    $sMyName, $sRemoteShell );
    return $iReturnCode; # No point in going ahead..
  }

  #-------------------------------------
  # Request clean up of bulletin board..
  #-------------------------------------

  #--------------------------------------------------
  # Attributes of interest for the active tuxmaster..
  #--------------------------------------------------
  my ( $sMstrRsrce, $sMstrEnvFile, $sMstrMcName, $sMstrMcType,
       $iAttempts, $sMstrTxDr, $sMstrTxDmn, $sMstrUsr,
       $iRetryPause, $sMstrSysName, @lsMstrAttr ) = ();

  $iAttempts = 0;   # Search attempts..
  $iRetryPause = 5; # Secs..

  while ( ( $iTimeout = $iTillWhen - time ) > 1 ) {
    &logTrace ( 20076,
    "<%s> +----- Searching for active master with [%s] sec. to spare -----+",
    $sMyName, $iTimeout );

    ++ $iAttempts; # Increment attempt counter..

    if ( Tuxedo::WhereIsMasterActive ( $sTuxDomain,
                                       $sTuxDir,
                                       \@lsMstrAttr ) ) {
      $iAttempts = 0; # Found! Reset..

      #--------------------------------------------------------
      # Get required attributes of the active master resource..
      #--------------------------------------------------------
      ( $sMstrSysName,
        $sMstrEnvFile,
        $sMstrMcName,
        $sMstrMcType,
        $sMstrTxDr,
        $sMstrTxDmn,
        $sMstrUsr,
        $sMstrRsrce ) = @lsMstrAttr;
  
      &logTrace ( 20077,
      "<%s> Found active master resource [%s] with:
      EnvFile     [%s]
      MachineName [%s]
      MachineType [%s]
      TuxDir      [%s]
      TuxDomain   [%s]
      User        [%s] online on
      VCS Node    [%s]",
      $sMyName,
      $sMstrRsrce, $sMstrEnvFile, $sMstrMcName, $sMstrMcType,
      $sMstrTxDmn, $sMstrTxDr, $sMstrUsr, $sMstrSysName );
  
      #-----------------------
      # Initialise variables..
      #-----------------------
      $iTimeout = $iTillWhen - time; # Time left..
      @lsCmdout = ();                # Truncate..
  
      &logTrace ( 20078,
      "<%s> Cleaning Tuxedo services for Tuxedo machine [%s]",
      $sMyName, $sMachineName );
  
      #---------------------------------------------------
      # Request TuxMaster to clean up the bulletin board..
      #---------------------------------------------------
      if ( Tuxedo::CleanBBofMachine ( $sMstrEnvFile,
                                      $sMstrTxDr,
                                      $sMachineName,
                                      $sMstrUsr,
                                      \@lsCmdout,
                                      $iTimeout,
                                      $sMstrSysName ) ) {
        &logTrace ( 20079,
        "<%s> Cleaned up bulletin board successfully.",
        $sMyName );
      }

      last; # Get out..
    }
    else {
      &logTrace ( 20080,
      "<%s> Could not locate active master in domain [%s] for attempt [%s]",
      $sMyName, $sTuxDomain, $iAttempts );

      #--------------------------------
      # Pause before attempting again..
      #--------------------------------
      sleep $iRetryPause; 
    }
  }

  if ( $iAttempts ) {
    &logInfo ( 20081,
    "<%s> Could not locate active master in domain [%s] in [%s] attempts.
    Returning [%s]", $sMyName, $sTuxDomain, $iAttempts, $iReturnCode );
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int GetResourceState ( IN String AppDir,
                                   IN String MachineName,
                                   IN String TuxServers,
                                   IN String User );

 Description : First level monitor to determine state of resource.

 Parameters :

 $sAppDir                   Absolute path to the Application directory
 $sMachineName              Logical Machine ID of this instance
 $sTuxServers               Names of the Tuxedo servers to monitor
 $sUser                     The dedicated UNIX login for this Tuxedo instance
 
 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 105  ONLINE with confidence level 50   VCSDef_MonitorExit::PARTIAL_ONLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub GetResourceState {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sAppDir,
       $sMachineName,
       $sTuxServers,
       $sUser ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20082, "Subroutine <%s> called:
  AppDir                   [%s]
  MachineName              [%s]
  TuxServers               [%s]
  User                     [%s]",

  $sMyName, $sAppDir, $sMachineName, $sTuxServers, $sUser );

  my @liDummyPids = ();
  my $iReturnCode = Tuxedo::GetProcState ( \@liDummyPids,
                                           $sAppDir,
                                           $sMachineName,
                                           $sTuxServers,
                                           $sUser,
                                           1 ); # Use FFLM..

  if ( VCSDef_MonitorExit::ONLINE == $iReturnCode ) {
    &logTrace ( 20083,
    "<%s> Found all required processes. Return code set to [%s]",
    $sMyName, $iReturnCode );
  }
  else {
    &logTrace ( 20084,
    "<%s> Did not find all required processes. Returning [%s]",
    $sMyName, $iReturnCode );
  }

  #--------------------------------
  # Populate the global PID array..
  #--------------------------------
  @liRunningProcs = @liDummyPids;
                                                                                
  &logTrace ( 20085,
  "<%s> Populating global running proclist with [%s] PIDs [%s]",
  $sMyName, scalar @liRunningProcs, join ( ' ', @liRunningProcs ) );

  return ( $iReturnCode );
}


=pod

 -------------------------------------------------------------------------
 Function : int GetProcState ( OUT ListReference OnlinePids,
                               IN String AppDir,
                               IN String MachineName,
                               IN String TuxServers,
                               IN String User,
                               IN Integer UseFFLM );

 Description : First level monitor to determine state of resource.

 Parameters :

 $rliOnlinePids             PIDs of relevant running processes
 $sAppDir                   Absolute path to the Application directory
 $sMachineName              Logical Machine ID of this instance
 $sTuxServers               Names of the Tuxedo servers to monitor
 $sUser                     The dedicated UNIX login for this Tuxedo instance
 $iUseFFLM                  Filtering procedure
 
 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 105  ONLINE with confidence level 50   VCSDef_MonitorExit::PARTIAL_ONLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub GetProcState {
  #----------------
  # Get Arguments..
  #----------------
  my ( $rliOnlinePids,
       $sAppDir,
       $sMachineName,
       $sTuxServers,
       $sUser,
       $iUseFFLM ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20086, "Subroutine <%s> called with:
  OnlinePids               [%s]
  AppDir                   [%s]
  MachineName              [%s]
  TuxServers               [%s]
  User                     [%s]
  UseFFLM                  [%s]",

  $sMyName, $rliOnlinePids, $sAppDir, $sMachineName,  $sTuxServers,
  $sUser, $iUseFFLM );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..

  #----------------------
  # Initialise Variables..
  #----------------------
  my ( $iProcessCount, $iUseRFLM, $iFound, $iUid, $iPid, $sLang,
       $sKey, $rhlProc, $sPidFile, $sProc, $sTemp,
       @hProcListHash, @lsUpdates, @lsValues,
       %hiProcPids, %hsSrvrDetails, %hsCopy, %hiAppHash, %hiOth ) = ();

  @$rliOnlinePids = ();
  
  #--------------------------
  # Other processes to kill..
  #--------------------------
  my @lsOthers = ( qr|.*\b$sMachineName\b.*$sAppDir\b| );

  #------------------------------------
  # Attempt to get monitoring details..
  #------------------------------------
  if ( ( 0 == ( $iUid = getpwnam $sUser ) ) && # User is the super-user..
       ( 2 == $iUseFFLM ) ) {                  # And called to kill-by-UID..
    &logTrace ( 20087, "<%s> Unsafe to kill processes by User [%s] UID [%s].
    Using filter-based approach.", $sMyName, $sUser, $iUid ),
    $iUseFFLM = 1;
  }

  if ( $iUseFFLM != 2 ) { # Called for first level check..
    if ( Tuxedo::GetSrvrDetails ( $sAppDir, 
                                  $sMachineName, 
                                  $sTuxServers, 
                                  \%hsSrvrDetails ) ) {
      &logTrace ( 20088,
      "<%s> Got details for monitoring Tuxedo instance", $sMyName );
  
      %hsCopy = %hsSrvrDetails;        # Backup..
      $iUseRFLM = scalar keys %hsCopy; # Set to entire count..
    }
    else {
      &logError ( 20089,
      "<%s> Failed to get details for monitoring Tuxedo instance",
      $sMyName );
      return VCSDef_MonitorExit::UNKNOWN;
    }
                                  
    #---------------------------------
    # Check if we attempt to be fast.. 
    #---------------------------------
    if ( $iUseFFLM ) { # FFLM..
      $iUseRFLM = 0;   # Hopefully skip..

      foreach $sKey ( keys %hsCopy ) {          # Use copy..
        @lsUpdates = ();                        # Reset..
        @lsValues = @{ $hsCopy { $sKey } };     # Get values..
        $sTemp .= "\nTuxedo process [$sKey]\n"; # Display string..

        #----------------
        # Read in pairs..
        #----------------
        while ( ( $sPidFile, $sProc ) = splice ( @lsValues, 0, 2 ) ) {

          #-----------
          # Run FFLM..
          #-----------
          $iFound = Tuxedo::FFLMStatus ( $sPidFile,
                                         $sProc,
                                         $iUid,
                                         \$iPid );

          #----------------
          # Update status..
          #----------------
          push ( @lsUpdates, $sPidFile, $sProc, $iFound );

          $sTemp .= "PID File     [$sPidFile]\n";
          $sTemp .= "Srch Process [$sProc]\n";
          $sTemp .= "FFLM Status  [$iFound]\n\n";
        }

        @{ $hsCopy { $sKey } } = @lsUpdates;

        #------------------------
        # Handle monitor status..
        #------------------------
        if ( VCSDef_MonitorExit::ONLINE == $iFound ) {
          push ( @$rliOnlinePids, $iPid ); # Push..
        }
        else {
          ++ $iUseRFLM; # Mark it for RFLM..
        }
      }

      #----------------
      # Print summary..
      #----------------
      &logTrace ( 20090,
      "<%s> Tuxedo instance process details ---%s\n\n",
      $sMyName, $sTemp );

      ( $sTemp, $sPidFile, $sProc ) = (); # Release memory..
    }
    else { # Use RLFM..
      #-----------------------------------------
      # Mark status of each service as offline..
      #-----------------------------------------
      $iFound = VCSDef_MonitorExit::OFFLINE; # Set to offline..
      foreach $sKey ( keys %hsCopy ) {
        @lsUpdates = (); # Reset..

        @lsValues = @{ $hsCopy { $sKey } }; # Get values..
        while ( scalar @lsValues ) {
          push ( @lsUpdates, splice ( @lsValues, 0, 2 ), $iFound );
        }

        @{ $hsCopy { $sKey } } = @lsUpdates; # Update..
      }
    }
  }
  else { # Called for Kill..
    &logTrace ( 20091,
    "<%s> Called to get processes for User [%s] UID [%s]",
    $sMyName, $sUser, $iUid );
    ++ $iUseRFLM;
  }

  #-----------------------
  # Regular slow :-( FLM..
  #-----------------------
  if ( $iUseRFLM ) {
    &logTrace ( 20092,
    "<%s> Running regular first level monitor", $sMyName );

    #----------------------------------
    # Set LANG for better performance..
    #----------------------------------
    $sLang = $ENV { LANG }; # Backup..
    $ENV { LANG } = 'C';    # Now set..

    #-----------------------------------
    # Call processes and store in hash..
    #-----------------------------------
    $iProcessCount = Proc::GetLongProcessListHash ( \@hProcListHash );

    #----------
    # Restore..
    #----------
    defined $sLang ? $ENV { LANG } = $sLang:delete $ENV { LANG }; # Restore..

    #----------------------
    # Check if we got any..
    #----------------------
    if ( $iProcessCount ) {
      &logTrace ( 20093,
      "<%s> Got [%s] processes from proc table for filtering",
      $sMyName, $iProcessCount );
    }
    else {
      &logError ( 20094,
      "<%s> Call to get process list did not return any processes",
      $sMyName );
      return VCSDef_MonitorExit::OFFLINE; # Offline..
    }

    #-----------------------
    # Search for processes..
    #-----------------------
    foreach $rhlProc ( @hProcListHash ) {
      #--------------------------------------------------------------
      # Each $rhlProc is a ref. to a hash with the following keys..
      # "UID"     =>  <user id>
      # "PID"     =>  <process id>
      # "PPID"    =>  <parent process id>
      # "COMMAND" =>  <command string>
      #--------------------------------------------------------------

      next unless defined $$rhlProc { UID }; # Bug in Proc::GetLongProcessListHash..
      next if $$rhlProc { PID } == $$;       # Skip self..
      $iPid = $$rhlProc { PID };             # Assign..

      &logTrace ( 20095, "<%s> Searching PID [%s] UID [%s] CMD [%s]",
      $sMyName, $$rhlProc { PID }, $$rhlProc { UID },
      $$rhlProc { COMMAND } );

      if ( $$rhlProc { UID } == $iUid ) { # Process owner..
        #-----------------------------------------------
        # Look for processes that were offline in FFLM..
        #-----------------------------------------------
        if ( $iUseFFLM != 2 ) { # Request to filter..
          foreach $sKey ( keys %hsCopy ) {
            @lsUpdates = (); # Reset..
            @lsValues = @{ $hsCopy { $sKey } }; # Get values..
    
            #--------------------------------------------------
            # Should have details per three elements of array..
            # <PIDFILE> <PROCESS> <STATE>
            #--------------------------------------------------
            while ( ( $sPidFile, $sProc, $iFound ) =
                      splice ( @lsValues, 0, 3 ) ) {
              if ( $$rhlProc { COMMAND } =~ /$sProc/ ) {        # Now found..
                unshift ( @$rliOnlinePids, $iPid );             # Store..
                if ( VCSDef_MonitorExit::OFFLINE == $iFound ) { # Earlier marked offline..
    
                  &logTrace ( 20096,
                  "<%s> Found PID [%s] PPID [%s] CMD [%s] with filter [%s]",
                  $sMyName, $iPid, $$rhlProc { PPID }, $$rhlProc { COMMAND },
                  $sProc );

                  $hiProcPids { $sKey } = $iPid;        # Store..
                  $iFound = VCSDef_MonitorExit::ONLINE; # Mark online..
                }
              }
    
              #----------------
              # Update status..
              #----------------
              push ( @lsUpdates, $sPidFile, $sProc, $iFound );
            }
    
            @{ $hsCopy { $sKey } } = @lsUpdates; # Update..
          }
        }
        else { # Request for kill..
          &logTrace ( 20097,
          "<%s> Pushing PID [%s] with PPID [%s] Cmd [%s] on stack",
          $sMyName, $iPid, $$rhlProc { PPID }, $$rhlProc { COMMAND } );
          push ( @$rliOnlinePids, $iPid ); # Store..
        } 
      }
      
      #---------------------------------
      # Handle other related processes..
      #---------------------------------
      if ( ( ! grep { /^$iPid$/ } @$rliOnlinePids ) &&           # Not caught earlier..
           $$rhlProc { UID } == $iUid ) {                        # Owned by User
        if ( map { $$rhlProc { COMMAND } =~ /$_/ } @lsOthers ) { # Was now found..
          push ( @$rliOnlinePids, $iPid );                       # Push..
          ++ $hiAppHash { $iPid };                               # And store..

          &logTrace ( 20098,
          "<%s> Matched process PID [%s] PPID [%s] CMD [%s]",
          $sMyName, $iPid, $$rhlProc { PPID }, $$rhlProc { COMMAND } );
        }
        else {
          $hiOth { $iPid } = $$rhlProc { PPID }; # Store the PPID..
        }
      }
    }

    #----------------------------------
    # Now get child processes to kill..
    #----------------------------------
    if ( scalar keys %hiAppHash && scalar keys %hiOth ) {
      push ( @$rliOnlinePids,
      grep { defined $hiAppHash { $hiOth { $_ } } } keys %hiOth );
    }

    ( $iPid, %hiAppHash, %hiOth ) = (); # Free..
  }

  #------------------------------------
  # Set final status for each process..
  #------------------------------------
  my ( $iOnline, $iOffline, $sEntryPointName ) = ();

  ( $sEntryPointName = $0 ) =~ s#.*/##; # Basename..
  $sTemp = ''; # Ensure..

  foreach $sKey ( keys %hsCopy ) {
    @lsValues = @{ $hsCopy { $sKey } }; # Get values..
    $sTemp .= "\nProcess      [$sKey]\n";

    while ( ( $sPidFile, $sProc, $iFound ) =
              splice ( @lsValues, 0, 3 ) ) {
      #-----------------------------------
      # Create PID files for use in FFLM..
      #-----------------------------------
      if ( $iUseRFLM                                 &&      # In RFLM..
           $iUseFFLM != 2                            &&      # Asked to filter..
           $hiProcPids { $sKey }                     &&      # PID populated..
           ( $sEntryPointName =~ /^(monitor|online)$/o ) ) { # In monitor or online..

        &logTrace ( 20099,
        "<%s> Creating PID file [%s] with PID [%s]",
        $sMyName, $sPidFile, $hiProcPids { $sKey } );

        #---------
        # Create..
        #---------
        if ( Tuxedo::CreatePidFile ( $sPidFile,
                                     $hiProcPids { $sKey } ) ) {
          &logTrace ( 20100,
          "<%s> Successfully created PID file [%s]", $sMyName, $sPidFile );
        }
        else {
          &logError ( 20101,
          "<%s> Could not create PID file [%s] successfully",
          $sMyName, $sPidFile );
        }
      }

      #-------------
      # Set status..
      #-------------
      if ( VCSDef_MonitorExit::ONLINE == $iFound ) {
        $iOnline = $iFound;
      }
      elsif ( VCSDef_MonitorExit::OFFLINE == $iFound ) {
        $iOffline = $iFound;
      }

      $sTemp .= "Srch Process [$sProc]\n";
      $sTemp .= "FLM Status   [$iFound]\n\n";
    }

    $sTemp .= "--------------------";
  }

  #-------------------
  # Set final status..
  #-------------------
  $iReturnCode = $iOffline || $iOnline || VCSDef_MonitorExit::UNKNOWN;

  #------------------------------------------------
  # Reset status based on PIDs, if not in monitor..
  #------------------------------------------------
  if ( $sEntryPointName ne 'monitor'                   &&
       ( VCSDef_MonitorExit::OFFLINE == $iReturnCode ) &&
       scalar @$rliOnlinePids ) {
    $iReturnCode = VCSDef_MonitorExit::PARTIAL_ONLINE;

    &logTrace ( 20102,
    "<%s> Found [%s] PIDs [%s] for this instance of Tuxedo.
    Resetting return code to [%s]",
    $sMyName, scalar @$rliOnlinePids, join ( ' ', @$rliOnlinePids ),
    $iReturnCode );
  }

  #-----------------------
  # Report on FLM status..
  #-----------------------
  if ( $iUseFFLM != 2 ) { # Request to filter..
    &logTrace ( 20103,
    "<%s> Tuxedo process details ---\n%s\nReturn code is [%s]",
    $sMyName, $sTemp, $iReturnCode );
    $sTemp = undef; # Release..
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int Monitor ( IN String EnvFile,
                          IN String MachineName,
                          IN String MachineType,
                          IN String MonitorProgram,
                          IN String RemoteShell,
                          IN Integer SecondLevelMonitor,
                          IN String TuxDir,
                          IN String TuxDomain,
                          IN String TuxServers,
                          IN String TuxServices,
                          IN String User );

 Description: Performs the in-depth monitoring

 Parameters :

 $sEnvFile                  Absolute path to environment file to source
 $sMachineName              Logical Machine ID of this instance
 $sMachineType              Bulletin board architecture
 $sMonitorProgram           Full path to external monitor program
 $sRemoteShell              Absolute path to rsh/ssh binary
 $iSecondLevelMonitor       Cycle count to run in-depth monitoring
 $sTuxDir                   Absolute path to Tuxedo base installation directory
 $sTuxDomain                Name of the Tuxedo domain for this instance
 $sTuxServers               Names of the Tuxedo servers to monitor
 $sTuxServices              Names of the Tuxedo services to monitor
 $sUser                     The dedicated UNIX login for this Tuxedo instance
 
 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub Monitor {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sEnvFile,
       $sMachineName,
       $sMachineType,
       $sMonitorProgram,
       $sRemoteShell,
       $iSecondLevelMonitor,
       $sTuxDir,
       $sTuxDomain,
       $sTuxServers,
       $sTuxServices,
       $sUser ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20104, "Subroutine <%s> called with:
  EnvFile                  [%s]
  MachineName              [%s]
  MachineType              [%s]
  MonitorProgram           [%s]
  RemoteShell              [%s]
  SecondLevelMonitor       [%s]
  TuxDir                   [%s]
  TuxDomain                [%s]
  TuxServers               [%s]
  TuxServices              [%s]
  User                     [%s]",

  $sMyName, $sEnvFile, $sMachineName, $sMachineType, $sMonitorProgram,
  $sRemoteShell, $iSecondLevelMonitor,  $sTuxDir, $sTuxDomain,
  $sTuxServers,  $sTuxServices, $sUser );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..
  my ( $bDelayedVal, $iTillWhen ) = ();

  my $sResourceName = $ENV { VCS_LOG_RESOURCE_NAME };

  #----------------------------------------------
  # In-depth monitoring, if correct time to run..
  #----------------------------------------------
  if ( VCSagentFW::RunSecondLevelMonitor ( $iSecondLevelMonitor ) ) {

    $iTillWhen = $^T + Tuxedo::GetEngineTimeout ( 1 );

    #--------------------------
    # Just In Time Validation..
    #--------------------------
    if ( VCSDef_MonitorExit::ONLINE == ( $iReturnCode =
         Tuxedo::DelayedAttributeCheck ( $sEnvFile,
                                         $sMachineName,
                                         $sMachineType,
                                         $sRemoteShell,
                                         $sTuxDir,
                                         $sTuxDomain,
                                         $sUser ) ) ) {
      &logTrace ( 20105,
      "<%s> Delayed attribute check completed successfully",
      $sMyName );

      ++ $bDelayedVal; # Set flag..
    }
    else {
      &logError ( 20106,
      "<%s> Attribute validation failed. Returning [%s]",
      $sMyName, $iReturnCode );
      return $iReturnCode;
    }

    &logTrace ( 20107,
    "<%s> Proceeding with second level monitoring", $sMyName );

    #-------------------------------------------------------
    # Upper bound..
    # Time in EPOCH when entry pt started + VCS time limit..
    #-------------------------------------------------------
    $iTillWhen ||= $^T + Tuxedo::GetEngineTimeout ( 1 );

    if ( VCSDef_MonitorExit::ONLINE == ( $iReturnCode =
         Tuxedo::SecondLevelMonitor ( $sEnvFile,
                                      $sMachineName,
                                      $sTuxDir,
                                      $sTuxServers,
                                      $sTuxServices,
                                      $sUser,
                                      $iTillWhen ) ) ) {

      &logTrace ( 20108,
      "<%s> Tuxedo instance online with status [%s] after SLM",
      $sMyName, $iReturnCode );
    }
    else {
      &logError ( 20109,
      "<%s> Tuxedo instance NOT online after SLM. Status [%s]",
      $sMyName, $iReturnCode );
      return $iReturnCode;
    }
  }

  #------------------------------------------------
  # Exit if external monitor has not been defined..
  #------------------------------------------------
  return $iReturnCode unless defined $sMonitorProgram;

  #---------------------------------------
  # MonitorProgram defined..
  # Separate the executable and the args..
  #---------------------------------------
  my ( $sExtCmd, $sExtCmdArgs ) = $sMonitorProgram # Separate
                                =~ /(\S+)\s*(.*)/; # Cmd and args..

  $sExtCmdArgs = '' unless defined $sExtCmdArgs;   # Prevent undefs..

  if ( $sExtCmd ) {                                # Defined..
    #----------------------------------------
    # Get time bearings, if not got earlier..
    #----------------------------------------
    $iTillWhen ||= $^T + Tuxedo::GetEngineTimeout ( 1 );

    if ( VCSDef_MonitorExit::ONLINE ==             # and found online..
         ( $iReturnCode =                          # by the status..
         Tuxedo::ExternalMonitor ( $sUser,
                                   $sEnvFile,
                                   $sExtCmd,
                                   $sExtCmdArgs,
                                   $iTillWhen ) ) ) {
      &logTrace ( 20110,
      "<%s> Tuxedo instance found online with status [%s]
      after executing external monitor [%s] with args [%s]",
      $sMyName, $iReturnCode, $sExtCmd, $sExtCmdArgs );
    }
    else {
      &logError ( 20111,
      "<%s> Tuxedo instance NOT online after executing external monitor
      [%s] with args [%s]. Status [%s]",
      $sMyName, $sExtCmd, $sExtCmdArgs, $iReturnCode );
    }
  }

  if ( $sMachineType eq 'Single' ||                   # SHM mode..
       $iReturnCode != VCSDef_MonitorExit::ONLINE ) { # Or not online..
    &logTrace ( 20112,
    "<%s> Monitoring for resource [%s] complete with status [%s]",
    $sMyName, $sResourceName, $iReturnCode );
    return $iReturnCode;
  }

  #-------------------------------
  # Monitoring an MP mode domain..
  #-------------------------------
  &logTrace ( 20113,
  "<%s> Proceeding with monitoring an MP mode domain.", $sMyName );

  #--------------------------
  # Just In Time Validation..
  #--------------------------
  if ( ! $bDelayedVal ) {
    if ( VCSDef_MonitorExit::ONLINE == ( $iReturnCode =
         Tuxedo::DelayedAttributeCheck ( $sEnvFile,
                                         $sMachineName,
                                         $sMachineType,
                                         $sRemoteShell,
                                         $sTuxDir,
                                         $sTuxDomain,
                                         $sUser ) ) ) {
      &logTrace ( 20114,
      "<%s> Delayed attribute check completed successfully",
      $sMyName );

      ++ $bDelayedVal; # Set flag..
    }
    else {
      &logError ( 20115,
      "<%s> Attribute validation failed. Returning [%s]",
      $sMyName, $iReturnCode );
      return $iReturnCode;
    }
  }

  if ( $iReturnCode != VCSDef_MonitorExit::ONLINE ) {
    return $iReturnCode;
  }

  #----------------------------------------
  # Get time bearings, if not got earlier..
  #----------------------------------------
  $iTillWhen ||= $^T + Tuxedo::GetEngineTimeout ( 1 );

  if ( Tuxedo::TuxedoMPModeMonitoring ( $sEnvFile,
                                        $sUser,
                                        $sTuxDir,
                                        $sTuxDomain,
                                        $sMachineName,
                                        $sMachineType,
                                        $sResourceName,
                                        $iTillWhen ) ) {
      #----------------------------------
      # Online in defined Tuxedo domain..
      #----------------------------------
      $iReturnCode = VCSDef_MonitorExit::ONLINE;

      &logTrace ( 20116,
      "<%s> Resource [%s] online with status [%s] in Tuxedo domain [%s]",
      $sMyName, $sResourceName, $iReturnCode, $sTuxDomain );
  }
  else {
    $iReturnCode = VCSDef_MonitorExit::OFFLINE;
    &logTrace ( 20117,
    "<%s> Resource [%s] offline with status [%s] in Tuxedo domain [%s]",
    $sMyName, $sResourceName, $iReturnCode, $sTuxDomain );
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int DelayedAttributeCheck ( IN String EnvFile,
                                        IN String MachineName,
                                        IN String MachineType,
                                        IN String RemoteShell,
                                        IN String TuxDir,
                                        IN String TuxDomain,
                                        IN String User );

 Description : Just In Time Validation

 Parameters :

 $sEnvFile                  Absolute path to environment file to source
 $sMachineName              Logical Machine ID of this instance
 $sMachineType              Bulletin board architecture
 $sRemoteShell              Absolute path to rsh/ssh binary
 $sTuxDir                   Absolute path to Tuxedo base installation directory
 $sTuxDomain                Name of the Tuxedo domain for this instance
 $sUser                     The dedicated UNIX login for this Tuxedo instance
 
 Returns  : Integer value as defined below:

 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub DelayedAttributeCheck {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sEnvFile,
       $sMachineName,
       $sMachineType,
       $sRemoteShell,
       $sTuxDir,
       $sTuxDomain,
       $sUser ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20118, "Subroutine <%s> called with:
  EnvFile                  [%s]
  MachineName              [%s]
  MachineType              [%s]
  RemoteShell              [%s]
  TuxDir                   [%s]
  TuxDomain                [%s]
  User                     [%s]",

  $sMyName,  $sEnvFile,  $sMachineName,  $sMachineType,
  $sRemoteShell, $sTuxDir, $sTuxDomain, $sUser );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..

  #--------------------------------------------
  # Create hash to weed out empty args values..
  #--------------------------------------------
  my @lsArgs = qw ( EnvFile MachineName MachineType RemoteShell TuxDir
                    TuxDomain User );

  my @lsReqdArgs = qw ( EnvFile MachineName MachineType TuxDir
                        TuxDomain User );

  my ( @lsMissing, %hsArgList ) = ();

  @hsArgList { @lsArgs } = @_; # Assign..

  #-----------------------
  # Check what's missing..
  #-----------------------
  if ( @lsMissing = grep { $hsArgList { $_ } !~ /\S/o } @lsReqdArgs ) {
    &logError ( 20119,
    "<%s> The following mandatory fields were not set [%s], Total [%s]",
    $sMyName, join ( ' ', @lsMissing ), scalar @lsMissing );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  #--------------------
  # Return, if failed..
  #--------------------
  if ( VCSDef_MonitorExit::ONLINE != $iReturnCode ) {
    return $iReturnCode;
  }

  #------------------------------
  # Validate parameter EnvFile..
  #------------------------------
  if ( ! -e $sEnvFile ) {
    &logError ( 20120, "<%s> [%s] file does not exist",
    $sMyName, $sEnvFile );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  #-----------------------------
  # Validate parameter TuxDir..
  #-----------------------------
  if ( ! -d $sTuxDir ) {
    &logError ( 20121, "<%s> [%s] directory does not exist",
    $sMyName, $sTuxDir );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  #---------------------------
  # Validate parameter User..
  #---------------------------
  if ( ! defined getpwnam $sUser ) {
    &logError ( 20122, "<%s> [%s] is not a valid UNIX login",
    $sMyName, $sUser );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  #-----------------------
  # Validate MachineType..
  #-----------------------
  if ( $sMachineType !~ /^((Non|)Master|Single|Backup)$/ ) {
    &logError ( 20123,
    "<%s> MachineType [%s] must be one of the following:
    Single, Master, NonMaster or Backup", $sMyName, $sMachineType );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Unknown..
  }

  return $iReturnCode if ( $sMachineType eq 'Single' ); # Over for SHM mode..

  #----------------------------------------
  # Verify rsh(1)/ssh(1) has been defined..
  #----------------------------------------
  if ( -x $sRemoteShell ) {
    $sRshCommand = $sRemoteShell; # Populate RSH_COMMAND..
  }
  else {
    &logError ( 20124,
    "<%s> If MachineType [%s] is not Single, RemoteShell [%s] has to be executable.",
    $sMyName, $sMachineType, $sRemoteShell );

    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Unknown..
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int SecondLevelMonitor ( IN String EnvFile,
                                     IN String MachineName,
                                     IN String TuxDir,
                                     IN String TuxServers,
                                     IN String TuxServices,
                                     IN String User,
                                     IN Integer TillWhen );

 Description : Does in-depth monitoring, if enabled.

 Parameters :

 $sEnvFile                  Absolute path to environment file to source
 $sMachineName              Logical Machine ID of this instance
 $sTuxDir                   Absolute path to Tuxedo base installation directory
 $sTuxServers               Names of the Tuxedo servers to monitor
 $sTuxServices              Names of the Tuxedo services to monitor
 $sUser                     The dedicated UNIX login for this Tuxedo instance
 $iTillWhen                 Time in EPOCH left for IInd level monitoring
 
 Returns  : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub SecondLevelMonitor {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sEnvFile,
       $sMachineName,
       $sTuxDir,
       $sTuxServers,
       $sTuxServices,
       $sUser,
       $iTillWhen ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20125, "Subroutine <%s> called with:
  EnvFile                  [%s]
  MachineName              [%s]
  TuxDir                   [%s]
  TuxServers               [%s]
  TuxServices              [%s]
  User                     [%s]
  TillWhen                 [%s]",

  $sMyName,  $sEnvFile,  $sMachineName,   $sTuxDir,   $sTuxServers,
  $sTuxServices, $sUser, $iTillWhen );  

  #------------------
  # Hopeful success..
  #------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE;

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sCommand, $sCommandArgs,
       $iOnlineCount, $iOfflineCount,$iTimeout, $iExitCode,
       @lsNeededServers, @lsNeededServices, @lsMissingProcs,
       @lsCmdout, %hiFilteredProcCount ) = ();

  #---------------------------------------------
  # Declare monitor command and args variables..
  #---------------------------------------------
  $sCommand = "$sTuxDir/bin/tmadmin";
  $sCommandArgs = "-r << EOF
psr -m $sMachineName
psc -m $sMachineName
quit
EOF
";

  #-------------------------------
  # Validate and execute command..
  #-------------------------------
  if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
    &logError ( 20126,
    "<%s> Insufficient time to continue in-depth monitoring.
    Short by [%s] sec",
    $sMyName, abs ( $iTimeout ) );

    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }
  elsif ( Tuxedo::ExecLocalCmd ( $sUser, 
                                 $sEnvFile, 
                                 $sCommand, 
                                 $sCommandArgs, 
                                 \$iExitCode, 
                                 $iTimeout, 
                                 \@lsCmdout ) ) {
    &logTrace ( 20127,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );
    
    #-------------------------------------------------
    # Filter application server/services by interest..
    #-------------------------------------------------
    grep {
      ( s/\n// || 1 )                 &&  # Chomp, and always make true..
      ! /^(Prog|Service) Name/o       &&  # Skip headers..
      /\b(\S+)\b/o                    &&  # Mark the server/service..
      ++ $hiFilteredProcCount { $1 }      # Store its count..
    } @lsCmdout;                          # from the output list..

    #-----------------------------------
    # Create list of requisite servers..
    #-----------------------------------
    @lsNeededServers = ( 'BBL', $sTuxServers =~ /([^(,|\s)]+)/g );

    #-------------------------------------------------
    # Check if we got at least one of what we wanted..
    #-------------------------------------------------
    if ( @lsMissingProcs =               # Store missing..
         grep {                          # When grep'ed..
           ! $hiFilteredProcCount { $_ } # And found missing..
         } @lsNeededServers ) {          # From what's needed..
      #------------
      # Log error..
      #------------
      &logError ( 20128,
      "<%s> Could not locate [%s] servers. Missing [%s]",
      $sMyName, join ( ' ', @lsMissingProcs ), scalar @lsMissingProcs );
  
      $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
    }
    else {
      &logTrace ( 20129,
      "<%s> Found all requested servers [%s]. Total [%s]",
      $sMyName, join ( ' ', @lsNeededServers ), scalar @lsNeededServers );
    }
  
    #--------------
    # Log summary..
    #--------------
    $iOfflineCount = scalar @lsMissingProcs;
    $iOnlineCount = scalar @lsNeededServers - scalar @lsMissingProcs;
  
    &logTrace ( 20130,
    "<%s> Total: Servers located [%s] Servers missing [%s] Return Code [%s]",
    $sMyName, $iOnlineCount, $iOfflineCount, $iReturnCode );
  
    #-----------------------------------------------
    # Return unless online and TuxServices defined..
    #-----------------------------------------------
    unless ( VCSDef_MonitorExit::ONLINE == $iReturnCode &&
             $sTuxServices =~ /\S/ ) {
      return $iReturnCode;
    }
    
    #-----------------------------------
    # Create list of requisite servers..
    #-----------------------------------
    @lsNeededServices = $sTuxServices =~ /([^(,|\s)]+)/g;
  
    #-------------------------
    # Now check the services..
    #-------------------------
    if ( @lsMissingProcs =            # Store missing..
         grep {                       # When grep'ed..
           ! $hiFilteredProcCount{$_} # And found missing..
         } @lsNeededServices ) {
      #------------
      # Log error..
      #------------
      &logError ( 20131,
      "<%s> Could not locate [%s] services. Missing [%s]",
      $sMyName, join ( ' ', @lsMissingProcs ), scalar @lsMissingProcs );
      $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Error..
    }
    else {
      &logTrace ( 20132,
      "<%s> Found all requested services [%s]. Total [%s]",
      $sMyName, join ( ' ', @lsNeededServices ), scalar @lsNeededServices );
    }
  
    #--------------
    # Log summary..
    #--------------
    $iOfflineCount = scalar @lsMissingProcs;
    $iOnlineCount = scalar @lsNeededServices - scalar @lsMissingProcs;
  
    &logTrace ( 20133,
    "<%s> Total: Services located [%s] Services missing [%s] Return code [%s]",
    $sMyName, $iOnlineCount, $iOfflineCount, $iReturnCode );
  }
  elsif ( ! defined $iExitCode ) {
    $lsCmdout [0] ||= 'undef'; # Set if not defined..
    $iReturnCode = $lsCmdout [0] eq 'timeout'? VCSDef_MonitorExit::TIMEOUT:
                                               VCSDef_MonitorExit::UNKNOWN; # Failure..
  }
  else {
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }

  return $iReturnCode;
}

=pod

 -------------------------------------------------------------------------
 Function :  bool ValidateMachineTypeInDomain ( IN String MachineType,
                                                IN String TuxDomain )

 Description : Verifies if, atmost one Master/Backup exists in given Domain,
               and the domain has either a Master or a Backup.
              
 Parameters :

 $sMachineType     Master, NonMaster, Backup or Single
 $sTuxDomain       Domain if MachineType is Master, NonMaster, Backup

 Returns: boolean value in the set of {0,1} where
          0 false
          1 true
 -------------------------------------------------------------------------

=cut

sub ValidateMachineTypeInDomain {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sMachineType,
       $sTuxDomain ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20134, "Subroutine <%s> called with:
  MachineType [%s]
  TuxDomain   [%s]", $sMyName, $sMachineType, $sTuxDomain );

  #-------------------------------------
  # Set return code to hopeful success..
  #-------------------------------------
  my $iReturnCode = 1;

  #------------------------------------
  # Return true if called with Single..
  #------------------------------------
  return $iReturnCode if ( $sMachineType eq 'Single' );

  #-------------------------
  # Define local variables..
  #-------------------------
  my $sCommand = $sCmdHares; # Get global..
  my $sCommandArgs = "-display -type Tuxedo " .
                     "-attribute MachineType TuxDomain";

  my ( $sMcType, $sDomain, $iExitCode, @lsCmdout, %hsMcTypes ) = ();

  #--------------------------------
  # Fire command and parse output..
  #--------------------------------
  @lsCmdout = grep {
                ! /^#/o                  && # Remove header..
                ( s/\n// || 1 )          && # Chomp.. make always true..
                /MachineType|TuxDomain/o && # If it contains one of these..
                s/.*\s+//                   # Get last field..
              } readpipe ( "$sCommand $sCommandArgs" );
  
  $iExitCode = $? >> 8; # Get exit status..

  #--------------------
  # Check exit status..
  #--------------------
  if ( $iExitCode ) {
    &logTrace ( 20135,
    "<%s> Command [%s] with Args [%s] failed with exitcode [%s]",
    $sMyName, $sCommand, $sCommandArgs, $iExitCode );
    return 0; # Error..
  }

  #-----------------------
  # Check for duplicates..
  #-----------------------
  while ( ( $sMcType, $sDomain ) = splice ( @lsCmdout, 0, 2 ) ) {
    if ( $sMcType =~ /^(Backup|Master)$/o      && # If Backup or Master..
         $hsMcTypes { $sMcType }               && # has been already found..
       ( $hsMcTypes { $sMcType } eq $sDomain ) && # in the same domain..
       ( $sTuxDomain eq $sDomain ) ) {            # when this domain concerns us..
      #----------------------
      # Configuration error..
      #----------------------
      &logError( 20136,
      "<%s> Configuration error. Duplicate MachineType: [%s] in Domain [%s]",
      $sMyName, $sMcType, $sDomain );
      $iReturnCode = 0; # Error..
    }
    else {
      $hsMcTypes { $sMcType } = $sDomain; # Store..
    }
  }

  #------------------------------------------------------
  # Check if we have at least one Master/Backup defined..
  #------------------------------------------------------
  unless ( $hsMcTypes { Master } || $hsMcTypes { Backup } ) {
    &logError( 20137,
    "<%s> Configuration error. At least one Master/Backup needed in Domain [%s]",
    $sMyName, $sTuxDomain );
    $iReturnCode = 0; # Error..
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int ExternalMonitor ( IN String User,
                                  IN String EnvFile,
                                  IN String Cmd,
                                  IN String CmdArgs,
                                  IN Integer TillWhen );

 Description : Handles user defined external monitor program

 Parameters :

 $sUser                     The dedicated UNIX login for this Tuxedo instance
 $sEnvFile                  Absolute path to environment file to source
 $sCmd                      User defined utilty to execute
 $sCmdArgs                  Command line arguments
 $iTillWhen                 Time in EPOCH left for IInd level monitoring
 
 Returns  : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub ExternalMonitor {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sUser,
       $sEnvFile,
       $sCmd,
       $sCmdArgs,
       $iTillWhen ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20138, "Subroutine <%s> called with:
  User                     [%s]
  EnvFile                  [%s]
  Cmd                      [%s]
  CmdArgs                  [%s]
  TillWhen                 [%s]",

  $sMyName, $sUser, $sEnvFile, $sCmd, $sCmdArgs, $iTillWhen );

  #------------------
  # Hopeful success..
  #------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE;

  #------------------------
  # Setup local variables..
  #------------------------
  my ( $iExitCode, $iTimeout, @lsCmdout ) = ();
  my $iBuf = 5; # Secs..

  #-------------------------
  # Handled exit code hash..
  #-------------------------
  my %hiExitCodes = (  0, VCSDef_MonitorExit::ONLINE,
                     110, VCSDef_MonitorExit::ONLINE,
                       1, VCSDef_MonitorExit::OFFLINE,
                     100, VCSDef_MonitorExit::OFFLINE,
                          VCSDef_MonitorExit::TIMEOUT,
                          VCSDef_MonitorExit::TIMEOUT );

  #---------------------------------------------
  # Check if command can be found and executed..
  #---------------------------------------------
  if ( ! -e $sCmd ) {
   &logError ( 20139,
    "<%s> External monitor [%s] does not exist",
    $sMyName, $sCmd );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( ! -x $sCmd ) {
   &logError ( 20140,
    "<%s> External monitor [%s] not executable.",
    $sMyName, $sCmd );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( $sEnvFile =~ /\S/ && ! -e $sEnvFile ) {
    &logError ( 20141,
    "<%s> EnvFile [%s] does not exist", $sMyName, $sEnvFile );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( ( $iTimeout = $iTillWhen - ( time + $iBuf ) ) <= 5 ) { # Lack of time..
   &logError ( 20142,
    "<%s> Insufficient time to proceed with external monitor [%s].
    Short by [%s] sec", $sMyName, $sCmd, abs ( $iTimeout ) );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }

  #--------------------------------------------
  # Return if validation failed at this point..
  #--------------------------------------------
  unless ( VCSDef_MonitorExit::ONLINE == $iReturnCode ) {
    return $iReturnCode;
  }

  &logTrace ( 20143,
  "<%s> Executing external monitor with Command [%s] Args [%s] in [%s] secs",
  $sMyName, $sCmd, $sCmdArgs, $iTimeout );

  #----------------------
  # Execute the command..
  #----------------------
  Tuxedo::ExecLocalCmd ( $sUser, 
                         $sEnvFile, 
                         $sCmd, 
                         $sCmdArgs, 
                         \$iExitCode, 
                         $iTimeout, 
                         \@lsCmdout );
                               
  #------------------------
  # Report on exit status..
  #------------------------
  if ( ! defined $iExitCode ) {
    $lsCmdout [0] ||= 'undef'; # Set if not defined..
    $iExitCode = $lsCmdout [0] eq 'timeout'? VCSDef_MonitorExit::TIMEOUT:
                                             VCSDef_MonitorExit::UNKNOWN; # Failure..
  }
  elsif ( $hiExitCodes { $iExitCode } &&
          VCSDef_MonitorExit::ONLINE ==
          $hiExitCodes { $iExitCode } ) {
    &logTrace ( 20144,
    "<%s> Command [%s] with args [%s] executed successfully with exit code [%s].
    Cmd output [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
  }
  else {
    &logWarn ( 20145,
    "<%s> Command [%s] with args [%s] execution failed with exit code [%s].
    Cmd output [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
  }

  #--------------------------------
  # Setup appropriate return code..
  #--------------------------------
  $iReturnCode = $hiExitCodes { $iExitCode } ||
                 VCSDef_MonitorExit::UNKNOWN;

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool CreatePidFile ( IN String PidFile,
                                 In Integer Pid );


 Description : Creates PID file with given PID

 Parameters :

 $sPidFile     File to create
 $iPid         PID to store

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub CreatePidFile {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sPidFile,
       $iPid ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20146, "Subroutine <%s> called with:
  PidFile         [%s]
  Pid             [%s]",

  $sMyName, $sPidFile, $iPid );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #-----------------------
  # Check against undefs..
  #-----------------------
  unless ( defined $iPid && $iPid =~ /^\d+$/o) {
    $iPid ||= 'undef'; # Prevent undefs.
    &logError ( 20147, "<%s> PID [%s] is not a valid positive number",
    $sMyName, $iPid );
    return 0; # Proceed no further..
  }

  #------------------
  # Create PID file..
  #------------------
  &logTrace ( 20148,
  "<%s> Creating PID file [%s] with PID [%s]",
  $sMyName, $sPidFile, $iPid );

  if ( sysopen ( PIDFILE, $sPidFile, O_WRONLY|O_SYNC|O_CREAT|O_TRUNC,
                 0600 ) ) {
    print PIDFILE $iPid, "\n"; # Update..

    unless ( close ( PIDFILE ) ) {
      &logWarn ( 20149,
      "<%s> Failed to close PidFile [%s] after update: [%s]",
      $sMyName, $sPidFile, $! );
      $iReturnCode = 0;
    }
  }
  else {
    &logWarn ( 20150,
    "<%s> Failed to create PID file [%s] with PID [%s]",
    $sMyName, $sPidFile, $iPid );
    $iReturnCode = 0;
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool DeletePidFile ( );


 Description : Deletes PID files for the VCS resource.

 Parameters :

 $sResourceName ResourceName for which the PID files needs to be deleted.

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
  Tuxedo::DeletePidFile ( );
 -------------------------------------------------------------------------

=cut

sub DeletePidFile {

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];
  my $sResourceName = $ENV { VCS_LOG_RESOURCE_NAME };

  &logTrace ( 20329, "Subroutine <%s> called with:
  ResourceName        [%s]",

  $sMyName, $sResourceName );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  my ( $sPidFileDir, $iRet ) = ();
  #-------------------
  # Delete PID files..
  #-------------------
  &logTrace ( 20330,
  "<%s> Deleting PID files for VCS resource [%s]",
  $sMyName, $sResourceName );

  $sPidFileDir = "/tmp/.VRTSTuxedo/" . $sResourceName . "/PID" ;

  if ( -d $sPidFileDir ) {
    $iRet = system("/bin/rm -rf $sPidFileDir") ;

    if ( $iRet == 0 ) {
      &logTrace ( 20331,
      "<%s> Successfully deleted PID files in the directory [%s].",
      $sMyName, $sPidFileDir );
      $iReturnCode = 0;
    }
    else {
      &logWarn ( 20332,
      "<%s> Failed to delete PID files in directory [%s]. Remove the files manually.",
      $sMyName, $sPidFileDir );
      $iReturnCode = 1;
    }
  }
  else {
    &logTrace ( 20333,
    "<%s> PID files for VCS resource [%s] doesn't exist.",
    $sMyName, $sResourceName );
  }

  return $iReturnCode;
}

=pod

 -------------------------------------------------------------------------
 Function : int FFLMStatus ( IN String PidFile,
                             IN String Proc,
                             IN Integer Uid,
                             OUT RefInt Pid );

 Description: Checks status of Process listed in PidFile, if running

 Parameters :

 $sPidFile       Absolute path to PID file of service
 $sProc          Process string to search for in proc table
 $iUid           UNIX UID of owner
 $riPid          Reference to existing PID

 Returns : Integer value as defined below:
 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 -------------------------------------------------------------------------

=cut

sub FFLMStatus {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sPidFile,
       $sProc,
       $iUid,
       $riPid ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20151, "Subroutine <%s> called with:
  PidFile                [%s]
  Proc                   [%s]
  UID                    [%s]
  riPid                  [%s]",

  $sMyName, $sPidFile, $sProc, $iUid, $riPid );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..
  $$riPid = 0; # Reset..

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $iPid, $rhlProc, @lProcessHash ) = ();

  #----------------------------------
  # Start by testing the attributes..
  #----------------------------------
  if ( $sProc !~ /\S/ ) {
    &logError ( 20152, "<%s> Called with empty [%s] process string",
    $sMyName, $sProc );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }
  elsif ( ! -e $sPidFile ) {
    &logTrace ( 20153, "<%s> PID file [%s] does not exist",
    $sMyName, $sPidFile );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }
  elsif ( ! -s $sPidFile ) {
    &logWarn ( 20154, "<%s> PID file [%s] is of zero bytes",
    $sMyName, $sPidFile );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }

  #---------------------
  # Return, if offline..
  #---------------------
  return $iReturnCode unless VCSDef_MonitorExit::ONLINE == $iReturnCode;
 
  if ( open ( PIDFILE, $sPidFile ) ) {
    ( $iPid ) = grep { ( s/\n// || 1 ) && # Chomp..
                         s/.*?(\d+)$/$1/  # Get PID..
                } <PIDFILE>;

    $iPid = '' unless defined $iPid;      # Avoid undefs..

    close ( PIDFILE ) ||                  # Close file..
    &logWarn ( 20155,
    "<%s> Could not close PID file [%s] after read [%s]",
    $sMyName, $sPidFile, $! );
  }
  else {
    &logError ( 20156,
    "<%s> Failed to open PID file [%s] for read: [%s]",
    $sMyName, $sPidFile, $! );
    return VCSDef_MonitorExit::OFFLINE; # Failure..
  }

  #------------------------
  # Check if PID is valid..
  #------------------------
  if ( $iPid =~ /\d+/ &&     # If the valid number..
       kill ( 0, $iPid ) ) { # is a PID that's alive..

    &logTrace ( 20157,
    "<%s> PID [%s] from PID file [%s] exists",
    $sMyName, $iPid, $sPidFile );

    #-----------------------
    # Search in proc table..
    #-----------------------
    if ( Tuxedo::GetPidDetails ( $iPid,
                                 \@lProcessHash ) ) {
      #--------------
      # Got details..
      #--------------
      foreach $rhlProc ( @lProcessHash ) {
        &logTrace ( 20158,
        "<%s> Process details: UID [%s] PID [%s] PPID [%s] ARGS [%s]",
        $sMyName, $$rhlProc { UID }, $$rhlProc { PID }, $$rhlProc { PPID },
        $$rhlProc { COMMAND } );

        #--------------
        # Now compare..
        #--------------
        if ( ( $$rhlProc { UID } == $iUid ) &&       # Owner match..
               $$rhlProc { COMMAND } =~ /$sProc/ ) { # Filter match..
          &logTrace ( 20159,
          "<%s> PID [%s] owned by UID [%s] with ARGS [%s] contains [%s]",
          $sMyName, $iPid, $$rhlProc { UID }, $$rhlProc { COMMAND }, $sProc );
          $$riPid = $iPid; # Store..
        }
        else {
          &logTrace ( 20160,
          "<%s> Ignoring PID [%s] owned by UID [%s] with ARGS [%s] with string [%s]",
          $sMyName, $iPid, $$rhlProc { UID }, $$rhlProc { COMMAND }, $sProc );
          $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
        }
      }
    }
    else {
      &logTrace ( 20161,
      "<%s> Could not get details for PID [%s]",
      $sMyName, $iPid );
      $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
    }
  }
  else {
    &logTrace ( 20162, "<%s> PID [$iPid] does not exist",
    $sMyName, $sPidFile );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }

  @lProcessHash = (); # Truncate..
  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool KillTuxedoProcesses ( IN String AppDir,
                                       IN String MachineName,
                                       IN String TuxServers,
                                       IN String User );

 Description  : Kills processes for this instance of Tuxedo
 
 Parameters :

 $sAppDir                 Absolute path to the Application directory
 $sMachineName            Logical Machine ID of this instance
 $sTuxServers             Names of the Tuxedo servers to monitor
 $sUser                   The dedicated UNIX login for this Tuxedo instance
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------
=cut

sub KillTuxedoProcesses {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sAppDir,
       $sMachineName,
       $sTuxServers,
       $sUser ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20163, "Subroutine <%s> called with:
  AppDir                   [%s]
  MachineName              [%s]
  TuxServers               [%s]
  User                     [%s]",

  $sMyName, $sAppDir, $sMachineName, $sTuxServers, $sUser );
 
  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1; # Success..

  #-------------------------
  # Set up local variables..
  #-------------------------
  my ( @liGoners, @liUnDead ) = ();

  #---------------------------------------------
  # Get list of processes running for this Uid..
  #---------------------------------------------
  if ( ( caller 1 )[3] eq 'Tuxedo::Start' ) { # During online..
    @liGoners = @liRunningProcs;              # Get PIDs to kill..
  }
  else {
    Tuxedo::GetProcState ( \@liGoners,
                           $sAppDir,
                           $sMachineName,
                           $sTuxServers,
                           $sUser,
                           0 ); # Set filtering choice..
  }

  #------------------------------------------------------------------
  # We do a two level kill. The SIGINQ ensures  that  the  PID  is  a
  # valid process before we actually issue the SIGKILL. We don't want
  # to report a failure for SIGKILL in case  of  defunct  or  missing
  # processes.
  #------------------------------------------------------------------
  if ( scalar @liGoners ) {
    #----------------------
    # Failed to kill some..
    #----------------------
    if ( @liUnDead = grep {
                         kill ( 0, $_ ) && # Valid..
                       ! kill ( 9, $_ )    # and could NOT be killed..
                     } @liGoners ) {       # From what's found..
      &logError ( 20164,
      "<%s> PIDs not killed [%s] Total not killed [%s]",
      $sMyName, join ( ' ', @liUnDead ), scalar @liUnDead );
      $iReturnCode = 0; # Failure..
    }
    else { # All killed..
      &logInfo ( 20165,
      "<%s> Killed PIDs [%s] Total killed: [%s]",
      $sMyName, join ( ' ', @liGoners ), scalar @liGoners );
    }
  }
  else {
    &logTrace ( 20166,
    "<%s> Could not locate any process this instance of Tuxedo",
    $sMyName );
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool ExecLocalCmd ( IN String User,
                                IN String EnvFile,
                                IN String Command,
                                IN String CommandArgs,
                                OUT RefInt ExitCode,
                                IN Integer Timeout,
                                OUT ListReference Cmdout );

 Description : Wrapper function for
               Sys::RunWithEnvCmdWithOutputWithTimeOut to collect output
               of the command.
 
 Parameters :

 $sUser            UNIX user to run the command as
 $sEnvFile         Absolute path to location of EnvFile to source
 $sCommand         Absolute path to location of command
 $sCommandArgs     Arguments to command
 $riExitCode       Reference to collect exit code of command
 $iTimeout         Time limit to execute the command
 $rlsCmdout        Stores o/p from command if successful
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub ExecLocalCmd {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sUser,
       $sEnvFile,
       $sCommand,
       $sCommandArgs,
       $riExitCode,
       $iTimeout,
       $rlsCmdout ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20167, "Subroutine <%s> called with:
  User        [%s]
  EnvFile     [%s]
  Command     [%s]
  CommandArgs [%s]
  riExitCode  [%s]
  Timeout     [%s]
  rlsCmdout   [%s]",

  $sMyName,
  $sUser, $sEnvFile, $sCommand, $sCommandArgs, $riExitCode,
  $iTimeout, $rlsCmdout );

  #---------------------------------------
  # Setup return code to hopeful success..
  #---------------------------------------
  my $iReturnCode = 1;

  #-------------------
  # Set up arguments..
  #-------------------
  my $bSourceUserEnv = 0; # Do not source..
  my $bTimedOut = 0;      # Timeout variable..

  @$rlsCmdout = (); # Reset..

  #-----------
  # Validate..
  #-----------
  if ( ! -e $sCommand ) {
    &logError ( 20168,
    "<%s> Cmd [%s] does not exist", $sMyName, $sCommand );
    $iReturnCode = 0; # Failure..
  }
  elsif ( ! -x $sCommand ) {
    &logError ( 20169,
    "<%s> Cmd [%s] is not executable", $sMyName, $sCommand );
    $iReturnCode = 0; # Failure..
  }
  elsif ( $iTimeout <= 1 ) {
    &logError ( 20170,
    "<%s> Insufficient time [%s] to execute Cmd [%s] with Args [%s].",
    $sMyName, $iTimeout, $sCommand, $sCommandArgs );
    $iReturnCode = 0; # Failure..
  }

  $iReturnCode || return $iReturnCode; # Return if failed.. 

  #----------------------
  # Set LOGNAME to user..
  #----------------------
  my $sOldLogName = $ENV { LOGNAME }; # Backup..
  $ENV { LOGNAME } = $sUser;          # Now set..

  &logTrace ( 20171,
  "<%s> Executing Command [%s] with Args [%s]",
  $sMyName, $sCommand, $sCommandArgs );

  $iReturnCode = Sys::RunWithEnvCmdWithOutputWithTimeOut ( $sEnvFile,
                                                           $bSourceUserEnv,
                                                           $sCommand,
                                                           $sCommandArgs,
                                                           $riExitCode,
                                                           $sUser,
                                                           $rlsCmdout,
                                                           \$bTimedOut,
                                                           $iTimeout );

  #---------------------------
  # Remove trailing newlines..
  #---------------------------
  chomp @$rlsCmdout;

  #--------------------
  # Check return code..
  #--------------------
  if ( $iReturnCode ) {   # No error with command execution..
    if ( $$riExitCode ) { # Failure..
      &logTrace ( 20172,
      "<%s> Command [%s] with args [%s] returned exit code [%s]",
      $sMyName, $sCommand, $sCommandArgs, $$riExitCode );
      $iReturnCode = 0; # Failure..
    }
    else {
      &logTrace ( 20173,
      "<%s> Command [%s] with args [%s] executed successfully with exit code [%s]",
      $sMyName, $sCommand, $sCommandArgs, $$riExitCode );
    }
  }
  else { # Execution error..
    @$rlsCmdout = (); # Reset..

    if ( $bTimedOut ) {  # We timed out..
      &logError ( 20174,
      "<%s> Command [%s] with args [%s] did not complete within [%s] seconds",
      $sMyName, $sCommand, $sCommandArgs, $iTimeout );
      $$rlsCmdout [0] = 'timeout'; # Timeout..
    }
    else { # Failure due to something else..
      &logError ( 20175,
      "<%s> Failed in call to RunWithEnvCmdWithOutputWithTimeOut",
      $sMyName );
    }
  }

  #----------
  # Restore..
  #----------
  defined $sOldLogName ? $ENV { LOGNAME } = $sOldLogName:delete $ENV { LOGNAME };

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool ExecRemoteTuxCmd ( IN String EnvFile,
                                    IN String User,
                                    IN String Command,
                                    IN String CommandArgs,
                                    IN RefernceList Cmdout,
                                    IN Integer Timeout,
                                    IN String SysName );

 Description : Executes given Tuxedo command on remote node
 
 Parameters :

 $sEnvFile         Absolute path to enviroment file for Tuxedo
 $sUser            UNIX user for Tuxedo instance
 $sCommand         Absolute path to location of shutdown command
 $sCommandArgs     Arguments to shutdown command
 $rlsCmdout        Stores o/p from command if successful
 $iTimeout         Time limit to execute the command
 $sSysName         VCS node name to run Command on

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub ExecRemoteTuxCmd {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sEnvFile,
       $sUser,
       $sCommand,
       $sCommandArgs,
       $rlsCmdout,
       $iTimeout,
       $sSysName ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20176, "Subroutine <%s> called with:
  EnvFile     [%s]
  User        [%s]
  Command     [%s]
  CommandArgs [%s]
  rlsCmdout   [%s]
  Timeout     [%s] to execute on
  SysName     [%s]",
  $sMyName, $sEnvFile, $sUser, $sCommand, $sCommandArgs,
  $rlsCmdout, $iTimeout, $sSysName );

  #---------------------------------------
  # Setup return code to hopeful success..
  #---------------------------------------
  my $iReturnCode = 1;

  #-------------------
  # Set up arguments..
  #-------------------
  my $rsRemoteShell = Tuxedo::RSH_COMMAND; # Get constant..
  my $bSourceUserEnv = 0; # Do not source..
  my ( $sUserShell, $sHome, $sCmdToRun, $bTimedOut,
       $iExitCode ) = ();

  @$rlsCmdout = (); # Reset..

  #------------------------------------
  # Get details about the remote user..
  #------------------------------------
  $sUser =~ s/^\+//; # Handle shared login..

  if ( ( $sUserShell, $sHome ) = ( getpwnam $sUser ) [-1, -2] ) {
    &logTrace ( 20177,
    "<%s> Using Shell [%s] Home [%s] Rsh command [%s] for remote User [%s]",
    $sMyName, $sUserShell, $sHome, $$rsRemoteShell, $sUser );
  }
  else {
    &logError ( 20178,
    "<%s> Could not get details for remote user [%s]", $sMyName, $sUser );
    return 0; # Error..
  }

  #---------------
  # Shell syntax..
  #---------------
  if ( $sUserShell =~ /csh$/ ) { # csh(1) based shell..
    &logTrace ( 20179,
    "<%s> User [%s] has csh derived native shell [%s]",
    $sMyName, $sUser, $sUserShell );
    $sCmdToRun = "setenv USER $sUser; setenv HOME $sHome;" .
    " umask 022; source $sEnvFile &&";
    $sCommandArgs =~ s/(\n|!)/\\$1/g; # Add trailing '\'s
  }
  else { # sh(1) based shell..
    &logTrace ( 20180,
    "<%s> User [%s] has sh derived native shell [%s]",
    $sMyName, $sUser, $sUserShell );
    $sCmdToRun = "USER=$sUser; HOME=$sHome; export USER HOME;" .
    " umask 022; . $sEnvFile &&";
  }

  #-------------------------------------
  # Recreate command for remote system..
  #-------------------------------------
  $sEnvFile = ""; # No local env..
  $sCommandArgs = "$sSysName \"$sCmdToRun $sCommand $sCommandArgs\"";
  $sCommand = $$rsRemoteShell;

  #-------------------------------------------------------
  # Disable X11 packet forwarding if dealing with ssh(1)..
  #-------------------------------------------------------
  $sCommandArgs = "-x $sCommandArgs" if ( $sCommand =~ /ssh.*/ );

  &logTrace ( 20181,
  "<%s> Executing [%s] with Args [%s] on [%s]",
  $sMyName, $sCommand, $sCommandArgs, $sSysName );

  $iReturnCode = Sys::RunWithEnvCmdWithOutputWithTimeOut ( $sEnvFile,
                                                           $bSourceUserEnv,
                                                           $sCommand,
                                                           $sCommandArgs,
                                                           \$iExitCode,
                                                           $sUser,
                                                           $rlsCmdout,
                                                           \$bTimedOut,
                                                           $iTimeout );
  #--------------------
  # Check return code..
  #--------------------
  if ( $iReturnCode ) { # No error with command execution..
    chomp @$rlsCmdout; # Chomp..

    #------------------------------------------------------------------
    # ssh(1) returns the exit code of the remote command. rsh(1)  never
    # does  so.   So a non-zero exit code for ssh(1) implies either the
    # ssh(1) or the remote command failed. For  rsh(1)  it  means  that
    # just the rsh(1) failed. rsh(1) would always return 0, even if the
    # remote command failed for any reason whatsoever.
    #------------------------------------------------------------------

    if ( $iExitCode ) { # Failed..
      &logTrace ( 20182, "<%s> Tuxedo Command [%s] returned exit code [%s]",
      $sMyName, $sCommand, $iExitCode );
      $iReturnCode = 0; # Failure..
      @$rlsCmdout = (); # Truncate..
    }
    else {
      &logTrace ( 20183, "<%s> Remote Shell [%s] executed successfully.",
      $sMyName, $sCommand );
    }
  }
  else { # Execution error..
    @$rlsCmdout = (); # Truncate..

    if ( $bTimedOut ) {  # We timed out..
      &logError ( 20184,
      "<%s> Tuxedo Command [%s] with args [%s] did not complete within [%s] seconds",
      $sMyName, $sCommand, $sCommandArgs, $iTimeout );
      $$rlsCmdout [0] = 'timeout'; # Timeout..
    }
    else { # Failure due to something else..
      &logError ( 20185, "<%s> Failed in call to RunCommandAsUserWithTimeout",
      $sMyName );
    }
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool GetPidDetails ( IN Integer Pid,
                                 OUT RefList ProcList );


 Description : Gets long process details for a known Pid if running

 Parameters :

 $iPid         PID whose details are requested
 $rlProcList   Ref to List to store details
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub GetPidDetails {
  #----------------
  # Get Arguments..
  #----------------
  my ( $iPid,
       $rlProcList ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20186, "Subroutine <%s> called with:
  Pid             [%s]
  rlProcList      [%s]",

  $sMyName, $iPid, $rlProcList );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #---------------
  # Validate PID..
  #---------------
  unless ( defined $iPid    &&   # Catch undefs..
           $iPid =~ /^\d+$/ &&   # And invalid PIDs..
           kill ( 0, $iPid ) ) { # And defunct ones..
    &logError ( 20187,
    "<%s> PID [%s] is invalid", $sMyName, $iPid );
    return 0;
  }

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sPsCmd, $sPsCmdArgs, $sLang, $sFmt, $sUnix95, $sSwitch, $iExitCode,
       %hsProc ) = ();

  $sSwitch = $^O;    # Get OS..
  @$rlProcList = (); # Truncate..

  #-------------
  # Store LANG..
  #-------------
  $sLang = $ENV { LANG };
  $ENV { LANG } = 'C';

  #---------------------------
  # Switch to the correct OS..
  #---------------------------

  if ( $sSwitch eq 'solaris') {
      $sPsCmd = "/usr/ucb/ps";
      $sPsCmdArgs = "axwwl $iPid";
      $sFmt = '[\da-z]+\s+([\d]+)\s+([\d]+)\s+([\d]+)\s+.*?[\d]:[\d]{2}\s(.*)';
  } elsif ( $sSwitch eq 'hpux' ) {
      #-----------------------
      # Force XPG4 behaviour..
      #-----------------------
      $sUnix95 = $ENV { UNIX95 };
      $ENV { UNIX95 } = 'XPG4';

      $sPsCmd = "/usr/bin/ps";
      $sPsCmdArgs = "-p $iPid -xxlo uid,pid,ppid,args";
      $sFmt = '([\d]+)\s+([\d]+)\s+([\d]+)\s+(.*)';
  } elsif ( $sSwitch eq 'aix' ) {
      $sPsCmd = "/usr/bin/ps";
      $sPsCmdArgs = "axwwl $iPid";
      $sFmt =  '[\da-e]+\s+[A-Z]\s+([\d]+)\s+([\d]+)\s+([\d]+)\s+.*?[\d]:[\d]{2}\s(.*)';
  } elsif ( $sSwitch eq 'linux' ) {
      $sPsCmd = "/bin/ps";
      $sPsCmdArgs = "wwo uid,pid,ppid,args $iPid";
      $sFmt = '([\d]+)\s+([\d]+)\s+([\d]+)\s+(.*)';
  }

  &logTrace ( 20188, "<%s> OS [%s] PsCmd [%s] PsCmdArgs [%s]",
  $sMyName, $sSwitch, $sPsCmd, $sPsCmdArgs );

  #-------------------
  # Validate command..
  #-------------------
  if ( ! -x $sPsCmd ) {
    &logError ( 20189,
    "<%s> ps(1b) command [%s] is not executable", $sMyName, $sPsCmd );
    return 0; # Failure..
  }

  #-------------------------------
  # Fire command and get details..
  #-------------------------------
  @hsProc { qw ( UID PID PPID COMMAND ) } =
  map {
    s/\n//;                             # Chomp..
    s/^\s+//;                           # Strip leading spaces..
    /^$sFmt$/o;                         # Extract fields..
  } readpipe ( "$sPsCmd $sPsCmdArgs" ); # From the command..

  $iExitCode = $? >> 8; # Get exit code..

  #------------------------
  # Restore env variables..
  #------------------------
  defined $sLang ? $ENV { LANG } = $sLang:delete $ENV { LANG };
  defined $sUnix95 ? $ENV { UNIX95 } = $sUnix95:delete $ENV { UNIX95 };

  #------------------
  # Check exit code..
  #------------------
  if ( $iExitCode ) {
    &logError ( 20190,
    "<%s> Cmd [%s] with args [%s] failed with exit code [%s]",
    $sMyName, $sPsCmd, $sPsCmdArgs, $iExitCode );

    %hsProc = (); # Truncate..
    return 0;
  }

  #---------------------
  # Fill up the bucket..
  #---------------------
  push ( @$rlProcList, \%hsProc ); # Update..

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool GetSrvrDetails ( IN String AppDir,
                                  IN String MachineName,
                                  IN String TuxServers,
                                  OUT RefHashList SrvrDetails );

 Description : Gives PID file and process string for given server type

 Parameters  :

 $sAppDir                   Absolute path to the Application directory
 $sMachineName              Logical Machine ID of this instance
 $sTuxServers               Names of the Tuxedo servers to monitor
 $rhlSrvrDetails            Stores details of PID file and process string
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub GetSrvrDetails {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sAppDir,
       $sMachineName,
       $sTuxServers,
       $rhlSrvrDetails ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20191, "Subroutine <%s> called with:
  AppDir                   [%s]
  MachineName              [%s]
  TuxServers               [%s]
  SrvrDetails              [%s]",

  $sMyName, $sAppDir, $sMachineName, $sTuxServers, $rhlSrvrDetails );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = 1; # Set to true..

  #-------------------
  # Define variables..
  #-------------------
  my ( $sKey, $sProc, $sPidFile, $sPIDDir, $sResourceType,
       $sResourceName ) = ();

  #-------------------------------------
  # List of search patterns to monitor..
  #-------------------------------------
  my @lsSrchPatterns = ( 'BBL', $sTuxServers =~ /([^(,|\s)]+)/g );

  #----------------------------------------------
  # Derive our resource type and resource names..
  #----------------------------------------------
  $sResourceType = ( $0 =~ m@([^/]+)@g ) [-2];
  $sResourceName = $ENV { VCS_LOG_RESOURCE_NAME };
  $sPIDDir = "/tmp/.VRTS$sResourceType/$sResourceName/PID";

  #------------------------------
  # Create PID folder if needed..
  #------------------------------
  if ( -d $sPIDDir ) {
    &logTrace ( 20192, "<%s> PID folder [%s] exists",
    $sMyName, $sPIDDir );
  }
  elsif ( ! mkpath ( $sPIDDir, 0, 0700 ) ) {
    &logWarn ( 20193, "<%s> Failed to create PID folder [%s]",
    $sMyName, $sPIDDir );
    return 0; # Cannot proceed..
  }

  %$rhlSrvrDetails = (); # Truncate..

  #-----------------------------------------------
  # Set appropriate filter strings and PID files..
  #-----------------------------------------------
  foreach $sKey ( @lsSrchPatterns ) {
    $sPidFile = "$sPIDDir/pidfile.$sKey";
    $sProc = qr|^$sKey\b\s.*\b$sMachineName\b.*$sAppDir\b|;
    @{ $$rhlSrvrDetails { $sKey } } = ( $sPidFile, $sProc ); # Store..
  }

  return $iReturnCode;
}

=pod

 -------------------------------------------------------------------------
 Function : bool WhereIsMasterActive ( IN String TuxDomain,
                                       IN String TuxDir,
                                       OUT StringRefernce TuxMstrAttr );

 Description  : Locates the resoure hosting the DBBL in the  given
                Tuxedo domain.

 In a given Tuxedo Domain, we can have atmost two TuxMasters,  one
 the Active Master (hosting the DBBL), the other the Backup Master
 (currently NOT hosting the DBBL). Only ONE can host the DBBL.

 We attempt to find  details  of  the  resource  within  the  same
 TuxDomain  that  is  currently  the  Active Master. We *MUST* get
 exactly one resource with  MachineType  "Master"  and  optionally
 exactly  one resource with MachineType "Backup". Anyone (but only
 ONE) can host the DBBL. Out of  these  two,  choose  those  whose
 State  is  ONLINE,  and check if it hosts the DBBL. If not, check
 the other.

 In order to reduce the remotely running tmadmin commands, we  try
 to  predict  the order to search these (at most) two resources by
 storing the details of the resource that  was  hosting  the  DBBL
 since  this  routine  ran for the specified TuxDomain, in a file.
 This file will be updated with the latest result on each run,  in
 order to predict which resource would be likely to host the DBBL.

 An incorrect order would cause the tmadmin  search  to  fail.  In
 this case, the file will be updated with the latest result.
 
 Parameters :

 $sTuxDomain      Domain if MachineType is Master, NonMaster, Backup
 $sTuxDir         Absolute path to location of Tuxedo instance
 $rlsTuxMstrAttr  Holds the following attributes of the active master
                  if found, in this order:

                  MstrSysName, MstrEnvFile, MstrMcName,
                  MstrMcType, MstrTxDr, MstrUsr, MstrRsrce

                  Else holds details of active backup in this order:
                  MstrSysName, MstrRsrce
      
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub WhereIsMasterActive {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sTuxDomain,
       $sTuxDir,
       $rlsTuxMstrAttr ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20194, "Subroutine <%s> called with:
  TuxDomain      [%s]
  TuxDir         [%s]
  rlsTuxMstrAttr [%s]",
  $sMyName, $sTuxDomain, $sTuxDir, $rlsTuxMstrAttr );

  #------------------------- 
  # Define local variables..
  #------------------------- 
  my $iReturnCode = 1; # Setup return code to hopeful success..
  my $iFound = 0;      # Flag if DBBL is found..
  my $sTuxSrchFile = "$sTuxDir/.ActiveMasterVCSInfo"; # History file..

  my ( $sMstrEnvFile,    $sMstrMcName,    $sMstrMcType,     $sMstrRsrce,
       $sMstrTxDr, $sMstrTxDmn, $sMstrUsr, $sMstrSysName, $i, $sCommand,
       $sCmdArgs,      $iExitCode,      @lsTuxHistory,      @lsHaresOut,
       @lsOnlineTuxBackup,  @lsOnlineTuxMaster,  @lsTemp, @lsFirstMatch,
       @lsSecondMatch ) = ();

  @$rlsTuxMstrAttr = (); # Clean up..

  #------------------------------------------------------------------
  # A history file holds the name of the resource last found  hosting
  # the  DBBL.   Hold these fields separated by '|' on a single line.
  # History file format:
  # <SysName>  <MstrEnvFile> <MstrMcName> <MstrMcType>
  # <MstrTxDr> <MstrTxDmn>   <MstrUsr>    <MstrRsrce>
  #------------------------------------------------------------------

  #---------------------
  # Check if its there..
  #---------------------
  if ( -e $sTuxSrchFile ) { # Allow a possible softlink..
    if ( sysopen ( FILDES, $sTuxSrchFile, O_RDONLY ) ) {
      #----------
      # Read in..
      #----------
      if ( 8 == ( ( $sMstrSysName, $sMstrEnvFile, $sMstrMcName, $sMstrMcType,
                    $sMstrTxDr, $sMstrTxDmn, $sMstrUsr, $sMstrRsrce ) = <FILDES>
                    =~ /([^|\n]+)/g ) ) {

        &logTrace ( 20195, "<%s> Found history file [%s] with following details:
        Last TuxMaster [%s]
        EnvFile        [%s]
        MachineName    [%s]
        MachineType    [%s]
        TuxDir         [%s]
        TuxDomain      [%s]
        User           [%s]
        SysName        [%s]",
        $sMyName,
        $sTuxSrchFile, $sMstrRsrce, $sMstrEnvFile, $sMstrMcName,
        $sMstrMcType, $sMstrTxDr, $sMstrTxDmn, $sMstrUsr, $sMstrSysName );

        #---------------------------------------
        # Attempt to get DBBL based on history..
        #---------------------------------------
        push ( @lsTuxHistory,
               $sMstrSysName, $sMstrEnvFile, $sMstrMcName,
               $sMstrMcType, $sMstrTxDr, $sMstrTxDmn, $sMstrUsr, $sMstrRsrce );

        push ( @lsOnlineTuxMaster, \@lsTuxHistory );

        #---------------------------------------------
        # Check for DBBL presence on above resources..
        #---------------------------------------------
        if ( $sMstrMcType =~ /^(Master|Backup)$/o && # Either a Backup or Master..
             $sTuxDomain eq $sMstrTxDmn           && # From our domain..
             Tuxedo::CheckDBBLOnResourceNode ( \@lsOnlineTuxMaster ) ) {
          &logTrace ( 20196,
          "<%s> Resource [%s] is online with active master [%s] in domain [%s]",
          $sMyName, $sMstrRsrce, $sMstrMcName, $sTuxDomain ); 
          push ( @$rlsTuxMstrAttr, @lsTuxHistory );
          return 1; # True..
        }
        else {
          &logTrace ( 20197,
          "<%s> Could not find active master for domain [%s] based on history file",
          $sMyName, $sTuxDomain );
        }
      }
      else { # The data is corrupt..
        ( $sMstrRsrce, $sMstrEnvFile, $sMstrMcName, $sMstrMcType,
          $sMstrTxDr, $sMstrUsr, $sMstrSysName ) = (); # Reset..
        &logTrace ( 20198,
        "<%s> Could not process data from history file [%s]",
         $sMyName, $sTuxSrchFile );
      }

      close ( FILDES ) ||
      &logError ( 20199,
      "<%s> Close failed after read on [%s]",
      $sMyName, $sTuxSrchFile );
    }
    else { # Open error..
      &logError ( 20200,
      "<%s> Open failed for read on [%s]", $sMyName, $sTuxSrchFile );
    }

    #------------------------------------------------------------------
    # Remove the incorrect file.. so that we don't refer to  it  again,
    # in case  we don't get DBBL at all with  the hares(1m) search  and
    # are called one more time..
    #------------------------------------------------------------------
    unlink ( $sTuxSrchFile ) ||
    &logError ( 20201,
    "<%s> Failed to unlink history file [%s]", $sMyName, $sTuxSrchFile );
  }

  #-------------------------------------------------------------------
  #  Since we are here.. we  do  not  have  a  history  file  to  rely
  #  on. Hence  we  now  query  the VCS, and then try to find the DBBL
  #  hosting resource out of the at most two likely  candidates..  Out
  #  of  these,  we can simply eliminate the one that's in the history
  #  file, since we've failed to find  DBBL  on  it.   This  remaining
  #  candidate  whose  resource  name is different from the one in the
  #  history file, *HAS* to have  a  MachineType  attribute  different
  #  from  that  in  the  history  file,  else we have a configuration
  #  error, where we have defined either two  Master  or  Backup  type
  #  resources in one domain!
  #-------------------------------------------------------------------

  &logTrace ( 20202,
  "<%s> Attempting an hares(1m) search for Master/Backup resource(s)",
  $sMyName );

  $sCommand = $sCmdHares; # Get global..
  $sCmdArgs = "-display -type Tuxedo -attribute " .
  "State EnvFile MachineName MachineType TuxDir TuxDomain User";

  #-----------------------------------------
  # Fire the command and collect the output..
  #-----------------------------------------
  @lsHaresOut =
  grep {
    ( s/\n// || 1 )                                   && # Chomp, and always make true..
    ( s/^(\S+).*\s(\S+)\s+ONLINE$/$1|$2/ && ++ $i )   || # If online, get composite, set flag..
    ( $i                                              && # If flag..
    /EnvFile|Machine(Name|Type)|TuxD(ir|omain)/o &&      # and it contains one of these..
    s/.*\s+// )                                       || # get the last field..
    ( $i && /User/                                    && # If flag, and has User..
    s/.*\s+/sprintf ("", -- $i )/e )                     # Get it, reset flag.
  } readpipe ( "$sCommand $sCmdArgs" );

  $iExitCode = $? >> 8; # Collect exit code..

  #------------------
  # Check exit code..
  #------------------
  if ( $iExitCode ) {
    &logError ( 20203,
    "<%s> Cmd [%s] with args [%s] failed with exit code [%s]",
    $sMyName, $sCommand, $sCmdArgs, $iExitCode );

    @lsHaresOut = (); # Truncate..
    return 0;         # Failure..
  }

  $i = 0;                               # Reset and reuse..
  $iFound = 0;                          # Reset.. ensure..
  ( @lsTemp, @lsOnlineTuxMaster ) = (); # Reset..

  if ( scalar @lsHaresOut ) {
    while ( scalar @lsHaresOut ) {
      @lsTemp = splice ( @lsHaresOut, 0, 7 ); # Extract in batches of 7..

      #---------------------------------------------
      # Get ResourceName, Node name from composite..
      #---------------------------------------------
      ( $lsTemp [7], $lsTemp [0] ) = $lsTemp [0] =~ /([^|]+)/g;
      $lsTemp [0] =~ s/.*://; # Remove GCO component from nodename..


      #-------------------------------------------------
      # Remove cluster component in name, if under GCO..
      #-------------------------------------------------
      $lsTemp [0] =~ s/.*://;

      &logTrace ( 20204,
      "<%s> VCS query report from [%s] with args [%s]:
      ------------------------------------
      Res. No. [%s]   Resource Name [%s]
                     EnvFile       [%s] 
                     MachineName   [%s] 
                     MachineType   [%s] 
                     TuxDir        [%s] 
                     TuxDomain     [%s] 
                     User          [%s]  online on
                     System        [%s]",
      $sMyName,
      $sCommand, $sCmdArgs, ++ $i,
      $lsTemp [7], $lsTemp [1], $lsTemp [2], $lsTemp [3],
      $lsTemp [4], $lsTemp [5], $lsTemp [6], $lsTemp [0] );

      #--------------------------------
      # Check if it is a likely match..
      #--------------------------------
      if ( $lsTemp [3] =~ /^(Master|Backup)$/o && # Either Master/Backup..
           $lsTemp [5] eq $sTuxDomain ) {         # From the same domain..
        ++ $iFound;                               # Raise flag..

      #-------------------------------------------------------------------
      #  We can find at most two candidates. If  we  have  data  from  the
      #  history  file  with  resource 'A' in it, and hares can return the
      #  following as online:
      #  i.  A and another resource B that can  host DBBL. In that case we
      #      neglect A.
      #  ii. 'A'. In that case we don't have anyone else. Error.
      #  iii.'B'. In that case we have just  'B' to  rely  on.
      #  iv. It returns a totally different set, 'B' & 'C'!
      #      In that case the history file has  been   tampered  with,  or
      #      is totally  unrealistic  for  the  current  setup.   In  that
      #      case, disregard the info in the history file and  investigate
      #      B & C.
      #-------------------------------------------------------------------

        if ( 3 == $iFound ) { # Oops!!
          ( @lsFirstMatch, @lsSecondMatch, @lsOnlineTuxMaster ) = (); # Clear..
          last;
        }

        #---------------------------------------
        # Store details of possible candidates..
        #---------------------------------------
        push ( @lsOnlineTuxBackup, $lsTemp [0], $lsTemp [7] );

        #--------------------------------------------
        # Compare with the data in the history file..
        #--------------------------------------------
        if ( join ( '|', @lsTuxHistory ) eq join ( '|', @lsTemp ) ) { # Skip..
          &logTrace ( 20205,
          "<%s> Skipping resource [%s] with MachineType [%s] based on history",
          $sMyName, $lsTemp [7], $lsTemp [3] ); 
          next;
        }

        #-----------------------------------------------------------------------
        # We are diffn. from one in the history file, or we don't have the file..
        #-----------------------------------------------------------------------
        &logTrace ( 20206,
        "<%s> An active master candidate has been located", $sMyName );

        #---------------------------
        # Store reference of match..
        #---------------------------
        if ( $lsTemp [3] eq 'Master' ) {
          @lsFirstMatch = @lsTemp;
          unshift ( @lsOnlineTuxMaster, \@lsFirstMatch ); # First place..
        }
        else { # Backup goes here..
          @lsSecondMatch = @lsTemp;
          push ( @lsOnlineTuxMaster, \@lsSecondMatch ); # Second place..
        }
      }
    }

    #------------------------
    # Check if we got lucky..
    #------------------------
    if ( scalar @lsOnlineTuxMaster ) { # Yup!!
      #-------------------------------------------
      # Verify that they have diff. MachineTypes..
      #-------------------------------------------
      if ( scalar @lsFirstMatch && scalar @lsSecondMatch ) {
        if ( $lsFirstMatch [3] eq $lsSecondMatch [3] ) {
          &logError ( 20207,
          "<%s> Identical MachineType [%s] found in domain [%s]",
          $sMyName, $lsFirstMatch [3], $sTuxDomain );
          $iReturnCode = 0; # Error..
        }   
      }
      else {
        &logTrace ( 20208,
        "<%s> Found [%s] possible candidates for checking DBBL",
        $sMyName, scalar @lsOnlineTuxMaster );
      }
    }
    else { # Oops!! 
      &logTrace ( 20209,
      "<%s> Could NOT find active master out of [%s] candidates. Returning immediately.",
      $sMyName, $iFound );
      $iReturnCode = 0; # Error..
    }
  }
  else { # Error!
    @$rlsTuxMstrAttr = (); # Clear..
    &logTrace ( 20210,
    "<%s> Found no ONLINE resources of type Tuxedo", $sMyName );
    $iReturnCode = 0; # Failure..
  }

  #-----------------------------------
  # Fill details of online resources..
  #-----------------------------------
  @$rlsTuxMstrAttr = @lsOnlineTuxBackup;

  $iReturnCode || return $iReturnCode; # Return if failed..

  #---------------------------------------------
  # Check for DBBL presence on above resources..
  #---------------------------------------------
  if ( $iReturnCode =
       Tuxedo::CheckDBBLOnResourceNode ( \@lsOnlineTuxMaster ) ) {
    &logTrace ( 20211,
    "<%s> Resource [%s] is the active master in domain [%s]",
    $sMyName, $lsOnlineTuxMaster [0]->[7], $sTuxDomain ); 

    $sMstrRsrce = $lsOnlineTuxMaster [0]; # Assign.. do NOT pop..
    @$rlsTuxMstrAttr = @$sMstrRsrce;      # Now dereference the array..

    #-----------------------------------------------
    # Update history file with the correct details..
    #-----------------------------------------------
    Tuxedo::UpdateTuxMasterHistory ( $sTuxDir,
                                     $rlsTuxMstrAttr );
  }
  else {
    &logTrace ( 20212,
    "<%s> Could not find active master for domain [%s]",
    $sMyName, $sTuxDomain );
  }

  return $iReturnCode;
}

=pod

 -------------------------------------------------------------------------
 Function Name: int GetMasterSearchWaitLimit ( IN String sMachineType,
                                               IN Integer iTillWhen );

 Description  : Calculates  timeout for  searching active master by
                MachineType based on current entry point  based  on
                our UNIX platform.

 Depending upon the  MachineType of this resource, the time  needed
 for it to declare, "A master cannot be found!", is determined.

 A MachineType  Master needs to come out earliest, to avoid a  race
 with  a Backup in case of  a cluster boot up. A NonMaster needs to
 be the slowest as it can be happy with anyone of above winning! :)

 Algorithm:
 ----------
 Timeleft ( secs )  = BaseTime - EPOCH
 Timeout ( secs )   = ( BaseTime - EPOCH ) * factor
 HuntTime ( EPOCH ) = ( ( BaseTime - EPOCH ) * factor ) + EPOCH

 
 Parameters :

 $sMachineType    Domain if MachineType is Master, NonMaster, Backup
 $iTillWhen       Time in EPOCH left for IInd level monitoring

 Returns : Time in EPOCH
 -------------------------------------------------------------------------

=cut

sub GetMasterSearchWaitLimit {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sMachineType,
       $iTillWhen ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20213, "Subroutine <%s> called with:
  MachineType   [%s]
  TillWhen      [%s]",
  $sMyName, $sMachineType, $iTillWhen );

  #-------------------
  # Get our platform..
  #-------------------
  my $sPlatform = $^O;

  #----------------------
  # Get our run context..
  #----------------------
  ( my $sEntryPoint = $0 ) =~ s#.*/##;

  #------------------------------------------
  # Define factors for lookup up Tux master..
  #------------------------------------------
  my $iFactor = 0; # Default..

  #----------------------------------------------------
  # We have purposely  separated ratios based on OS, in
  # case we want to change them in the future if needed
  #----------------------------------------------------

  #------------------------------------------
  # Solaris. Based on MonitorTimeout >= 300..
  #------------------------------------------
  my %hiMstrSrchFactorCleanSolaris = ( 'Master',    1.0,
                                       'Backup',    1.0,
                                       'NonMaster', 1.0 );
  
  my %hiMstrSrchFactorMonitorSolaris = ( 'Master',    0.1,
                                         'Backup',    0.1,
                                         'NonMaster', 0.3 );
  
  my %hiMstrSrchFactorOnlineSolaris = ( 'Master',    0.1,
                                        'Backup',    0.5,
                                        'NonMaster', 0.9 );
  
  my %hiMstrSrchFactorOfflineSolaris = ( 'Master',    0.4,
                                         'Backup',    0.3,
                                         'NonMaster', 0.2 );

  my %hsrTimingsSolaris = ( 'clean',   \%hiMstrSrchFactorCleanSolaris,
                            'monitor', \%hiMstrSrchFactorMonitorSolaris,
                            'online',  \%hiMstrSrchFactorOnlineSolaris,
                            'offline', \%hiMstrSrchFactorOfflineSolaris );

  #----------------------------------------
  # Linux. Based on MonitorTimeout >= 300..
  #----------------------------------------
  my %hiMstrSrchFactorCleanLinux = ( 'Master',    1.0,
                                     'Backup',    1.0,
                                     'NonMaster', 1.0 );

  my %hiMstrSrchFactorMonitorLinux = ( 'Master',    0.1,
                                       'Backup',    0.1,
                                       'NonMaster', 0.3 );

  my %hiMstrSrchFactorOnlineLinux = ( 'Master',    0.1,
                                      'Backup',    0.5,
                                      'NonMaster', 0.9 );

  my %hiMstrSrchFactorOfflineLinux = ( 'Master',    0.4,
                                       'Backup',    0.3,
                                       'NonMaster', 0.2 );

  my %hsrTimingsLinux = ( 'clean',   \%hiMstrSrchFactorCleanLinux,
                          'monitor', \%hiMstrSrchFactorMonitorLinux,
                          'online',  \%hiMstrSrchFactorOnlineLinux,
                          'offline', \%hiMstrSrchFactorOfflineLinux );

  #---------------------------------------
  # HPUX. Based on MonitorTimeout >= 300..
  #---------------------------------------
  my %hiMstrSrchFactorCleanHPUX = ( 'Master',    1.0,
                                    'Backup',    1.0,
                                    'NonMaster', 1.0 );

  my %hiMstrSrchFactorMonitorHPUX = ( 'Master',    0.1,
                                      'Backup',    0.1,
                                      'NonMaster', 0.3 );

  my %hiMstrSrchFactorOnlineHPUX = ( 'Master',    0.1,
                                     'Backup',    0.5,
                                     'NonMaster', 0.9 );

  my %hiMstrSrchFactorOfflineHPUX = ( 'Master',    0.4,
                                      'Backup',    0.3,
                                      'NonMaster', 0.2 );

  my %hsrTimingsHPUX = ( 'clean',   \%hiMstrSrchFactorCleanHPUX,
                         'monitor', \%hiMstrSrchFactorMonitorHPUX,
                         'online',  \%hiMstrSrchFactorOnlineHPUX,
                         'offline', \%hiMstrSrchFactorOfflineHPUX );

  #--------------------------------------
  # AIX. Based on MonitorTimeout >= 300..
  #--------------------------------------
  my %hiMstrSrchFactorCleanAix = ( 'Master',    1.0,
                                   'Backup',    1.0,
                                   'NonMaster', 1.0 );

  my %hiMstrSrchFactorMonitorAix = ( 'Master',    0.1,
                                     'Backup',    0.1,
                                     'NonMaster', 0.3 );

  my %hiMstrSrchFactorOnlineAix = ( 'Master',    0.1,
                                    'Backup',    0.5,
                                    'NonMaster', 0.9 );

  my %hiMstrSrchFactorOfflineAix = ( 'Master',    0.4,
                                     'Backup',    0.3,
                                     'NonMaster', 0.2 );

  my %hsrTimingsAix = ( 'clean',   \%hiMstrSrchFactorCleanAix,
                        'monitor', \%hiMstrSrchFactorMonitorAix,
                        'online',  \%hiMstrSrchFactorOnlineAix,
                        'offline', \%hiMstrSrchFactorOfflineAix );

  #-------------
  # Store them..
  #-------------
  my %hrhiEntryPointsUNIX = ( 'aix',     \%hsrTimingsAix,
                              'hpux',    \%hsrTimingsHPUX,
                              'linux',   \%hsrTimingsLinux,
                              'solaris', \%hsrTimingsSolaris );

  #---------------------
  # Check if we got it..
  #---------------------
  if ( $hrhiEntryPointsUNIX { $sPlatform } ) {
    $iFactor = $hrhiEntryPointsUNIX { $sPlatform }     # Populate..
                                 -> { $sEntryPoint }
                                 -> { $sMachineType };

    &logTrace ( 20214,
    "<%s> Platform [%s] Entry Point [%s] MachineType [%s] Factor [%s]",
    $sMyName, $sPlatform, $sEntryPoint, $sMachineType, $iFactor );
  }
  else {
    &logError ( 20215,
    "<%s> No data for MachineType [%s] Entry Point [%s] on Platform [%s]",
    $sMyName, $sMachineType, $sEntryPoint, $sPlatform );
    return $iFactor; # Error. Return default..
  }

  #--------------------------------------------
  # Return search time for master in EPOCH..
  #--------------------------------------------
  my $iNow = time; # EPOCH..
  my $iHuntTime = ( ( $iTillWhen - $iNow ) * $iFactor ) + $iNow;

  &logTrace ( 20216,
  "<%s> Setting MasterSearchWaitLimit to [%s] EPOCH secs",
  $sMyName, $iHuntTime );

  return ( $iHuntTime );
}


=pod

 -------------------------------------------------------------------------

 Function :  bool TuxedoMPModeMonitoring ( IN String EnvFile,
                                           IN String User,
                                           IN String TuxDir,
                                           IN String TuxDomain,
                                           IN String MachineName,
                                           IN String MachineType,
                                           IN String ResourceName,
                                           IN Integer TillWhen );

 Description : Monitoring logic for MP mode of Tuxedo

 Parameters :

 $sEnvFile        Absolute path to enviroment file for Tuxedo
 $sUser           UNIX user for Tuxedo instance
 $sTuxDir         Absolute path to location of Tuxedo instance
 $sTuxDomain      Domain if MachineType is Master, NonMaster, Backup
 $sMachineName    Machine name registered with Tuxedo
 $sMachineType    Master, NonMaster, Backup or Single
 $sResourceName   Name of this resource
 $iTillWhen       Timeout in secs for completing activity

 Returns: boolean value in the set of {0,1} where
          0 false
          1 true
 -------------------------------------------------------------------------

=cut

sub TuxedoMPModeMonitoring {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sEnvFile,
       $sUser,
       $sTuxDir,
       $sTuxDomain,
       $sMachineName,
       $sMachineType,
       $sResourceName,
       $iTillWhen ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20217, "Subroutine <%s> called with:
  EnvFile      [%s]
  User         [%s]
  TuxDir       [%s]
  TuxDomain    [%s]
  MachineName  [%s]
  MachineType  [%s]
  ResourceName [%s]
  TillWhen     [%s]",
  $sMyName,
  $sEnvFile, $sUser, $sTuxDir, $sTuxDomain, $sMachineName,
  $sMachineType, $sResourceName, $iTillWhen );

  #------------------
  # Hopeful success..
  #------------------
  my $iReturnCode = 1;
  my $sNodeName = ( uname )[1]; # Grab second element..

  #--------------------------------------------------
  # Attributes of interest for the active tuxmaster..
  #--------------------------------------------------
  my ( $sMstrRsrce, $sMstrEnvFile, $sMstrMcName, $iTimeout,
       $sMstrMcType, $sMstrTxDr, $sMstrTxDmn, $sMstrUsr,
       $sMstrSysName, @lsMstrAttr ) = (); 

  my ( $sCommand, $sCommandArgs ) = ();

  #---------------------------------------
  # Find the active master in the domain..
  #---------------------------------------
  my $iAttempts = 0;   # Search attempts..
  my $iRetryPause = 5; # Seconds..
  my $iMstrSrchLmt = Tuxedo::GetMasterSearchWaitLimit ( $sMachineType,
                                                        $iTillWhen );

  while ( ( $iTimeout = $iMstrSrchLmt - time ) > 1 ) {

    ++ $iAttempts; # Increment attempt counter..

    #-------------------------------
    # Search for the active master..
    #-------------------------------
    &logTrace ( 20218,
    "<%s> +----- Searching for active master with [%s] sec. to spare -----+",
    $sMyName, $iTimeout );

    if ( Tuxedo::WhereIsMasterActive ( $sTuxDomain,
                                       $sTuxDir,
                                       \@lsMstrAttr ) ) {
      #--------------------------------------------------------
      # Get required attributes of the active master resource..
      #--------------------------------------------------------
      ( $sMstrSysName,
        $sMstrEnvFile,
        $sMstrMcName,
        $sMstrMcType,
        $sMstrTxDr,
        $sMstrTxDmn,
        $sMstrUsr,
        $sMstrRsrce ) = @lsMstrAttr;

      &logTrace ( 20219, "<%s> Found active master resource [%s] with:
      EnvFile     [%s]
      MachineName [%s]
      MachineType [%s]
      TuxDir      [%s]
      TuxDomain   [%s]
      User        [%s] online on
      VCS Node    [%s]",
      $sMyName, $sMstrRsrce, $sMstrEnvFile, $sMstrMcName, $sMstrMcType,
      $sMstrTxDr, $sMstrTxDmn, $sMstrUsr, $sMstrSysName );

      $iAttempts = 0; # Reset to indicate success..
      last;           # Found.. get out..
    }
    else {
      &logTrace ( 20220,
      "<%s> Failed to find active Master in domain [%s] in attempt [%s]",
      $sMyName, $sTuxDomain, $iAttempts );

      #----------------------------
      # Inform the backup machine..
      #----------------------------
      if ( 1 == $iAttempts              && # Only once..
           $sMachineType eq 'NonMaster' && # If NonMaster..
           scalar @lsMstrAttr ) {          # We have the backup details..
        #------------------------------------
        # Get details about the backup node..
        #------------------------------------
        while ( ( $sMstrSysName, $sMstrRsrce ) = splice ( @lsMstrAttr, 0, 2 ) ) {
          &logTrace ( 20221,
          "<%s> Probing resource [%s] on system [%s] to expedite master migration",
          $sMyName, $sMstrRsrce, $sMstrSysName );
  
          $sCommand = $sCmdHares; # Get global..
          $sCommandArgs = "-probe $sMstrRsrce -sys $sMstrSysName";
          system ( "$sCommand $sCommandArgs" ); # Probe.. Ignore return code..
        }
      }

      sleep $iRetryPause; # Pause, before trying again..
    }
  }

  #----------------------------------
  # We failed to find active master..
  #----------------------------------
  if ( $iAttempts ) {
    #----------
    # Migrate..
    #----------
    if ( $sMachineType =~ /^(Master|Backup)$/o ) {
      &logWarn ( 20222,
      "<%s> Could not locate active Master for domain [%s] after [%s] attempts.
      ATTEMPTING TO MIGRATE master",
      $sMyName, $sTuxDomain, $iAttempts );

      #--------------------
      # Get time bearings..
      #--------------------
      if ( ( $iTimeout = $iTillWhen - time ) <= 1 ) {
        &logError ( 20223,
        "<%s> Insufficient time to migrate master. Short by [%s] sec",
        $sMyName, abs ( $iTimeout ) );
        return 0; # Failure..
      }

      if ( Tuxedo::MigrateTuxedoMaster ( $sTuxDir,
                                         $sEnvFile,
                                         $sUser,
                                         $sMachineName,
                                         $iTimeout ) ) {
        &logInfo ( 20224,
        "<%s> Master migration procedure successful.
        Machine [%s] is the new active master", $sMyName, $sMachineName );

        #----------------------------------------------
        # Update history file for WhereIsMasterActive..
        #----------------------------------------------
        @lsMstrAttr = ( $sNodeName, $sEnvFile, $sMachineName,
                        $sMachineType, $sTuxDir, $sUser,
                        $sResourceName );

        Tuxedo::UpdateTuxMasterHistory ( $sTuxDir,
                                         \@lsMstrAttr );
      }
      else {
        &logError ( 20225,
        "<%s> Migration FAILED! Quitting!", $sMyName );

        $iReturnCode = 0; # Failure..
      }
    }
    else { # NonMaster..
      &logError ( 20226,
      "<%s> Could not locate active Master for domain [%s] after [%s] attempts.
      This [%s] machine is PARTITIONED!",
      $sMyName, $sTuxDomain, $iAttempts, $sMachineType );

      $iReturnCode = 0; # Failure..
    }
  }
  else { # Active master found..
    #------------------------------------------
    # Verify that we aren't the active master..
    #------------------------------------------
    if ( $sMstrMcName ne $sMachineName ) { # We are not..
      #-----------------------------
      # Get time bearings in EPOCH..
      #-----------------------------
      $iMstrSrchLmt = Tuxedo::GetEngineTimeout ( 0 ) # Without grace limit..
                      - 15 + $^T;                    # Setting grace here..

      #-----------------------------------
      # Start the search for our identity!
      #-----------------------------------
      $iAttempts = 0;   # Reset..
      $iRetryPause = 5; # Seconds..

      while ( ( $iTimeout = ( $iMstrSrchLmt - time ) ) > 1 ) {

        ++ $iAttempts; # Increment search counter..

        &logTrace ( 20227,
        "<%s> +----- Getting domain membership with [%s] sec remaining -----+",
        $sMyName, $iTimeout );

        if ( Tuxedo::GetDomainMembershipStatus ( $sUser,
                                                 $sEnvFile,
                                                 $sTuxDir,
                                                 $sTuxDomain,
                                                 $sMachineName,
                                                 $sMachineType,
                                                 $sResourceName,
                                                 $iAttempts,
                                                 $iTimeout ) ) {
          $iAttempts = 0; # Indicate success..

          &logTrace ( 20228,
          "<%s> Found valid BBL with Master [%s]. Machine [%s] is part of TuxDomain [%s]",
          $sMyName, $sMstrMcName, $sMachineName, $sTuxDomain );
          last; # Get out..
        }
        else {
          &logTrace ( 20229,
          "<%s> Could NOT get validated by the active master [%s] after [%s] attempts",
          $sMyName, $sMstrMcName, $iAttempts );

          sleep $iRetryPause; # Pause, before trying again..
        }
      }

      #-------------------------------
      # Report if we are partitioned..
      #-------------------------------
      if ( $iAttempts ) {
        &logError ( 20230,
        "<%s> Could NOT locate machine's BBL with the active master [%s] after [%s] attempts.
        MACHINE IS PARTITIONED", $sMyName, $sMstrMcName, $iAttempts );

        $iReturnCode = 0; # Error..
      }
    }
    else { # We are the active master..
      &logTrace ( 20231,
      "<%s> THIS is the active master in [%s] domain.", $sMyName, $sTuxDomain ); 
    }
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool UpdateTuxMasterHistory ( IN String TuxDir,
                                          IN ListRefernce TuxMstrAttr );

 Description : Updates history file for WhereIsMasterActive

 Parameters :

 $sTuxDir                   Absolute path to Tuxedo base installation directory
 $rlsTuxMstrAttr            Details regarding 

 History file has following fields on a SINGLE line separated by '|'

 <SysName> <MstrEnvFile> <MstrMcName> <MstrMcType>
 <MstrTxDr> <MstrTxDmn> MstrUsr> <MstrRsrce>

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut


sub UpdateTuxMasterHistory {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sTuxDir,
       $rlsTuxMstrAttr ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20232, "Subroutine <%s> called with:
  TuxDir         [%s]
  rlsTuxMstrAttr [%s]",
  $sMyName, $sTuxDir, $rlsTuxMstrAttr );

  #-------------------------------------
  # Set return code to hopeful success..
  #-------------------------------------
  my $iReturnCode = 1;

  #-----------------------
  # History file details..
  #-----------------------
  my $sHisFile = "$sTuxDir/.ActiveMasterVCSInfo";

  #-----------------------------------------------
  # Update history file with the correct details..
  #-----------------------------------------------
  if ( sysopen ( FILDES,
       $sHisFile, O_WRONLY|O_CREAT|O_TRUNC|O_SYNC, 0600 ) ) {
    &logTrace ( 20233, "<%s> Opened history file [%s] for write",
    $sMyName, $sHisFile );

    print FILDES join ( '|', @$rlsTuxMstrAttr ), "\n"; # Update..

    close ( FILDES ) ||
    &logError ( 20234,
    "<%s> Failed to close history file [%s] after write [%s]",
    $sMyName, $sHisFile, $! );
  }
  else {
    &logError ( 20235,
    "<%s> Failed to open history file [%s] for write [%s]",
    $sMyName, $sHisFile, $! );

    $iReturnCode = 0; # Error..
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool CheckDBBLOnResourceNode ( IN/OUT
                                           RefListRefList TuxResources );

 Description : Searches DBBL for given resources and updates parameter
               with details of resource, else trucates it, on failure.
 
 Parameters :

 $rlrlsTuxResources     Holds ref. to a list that has ref.
                        of lists with these string elements:
                        [0] <SysName>
                        [1] <MstrEnvFile>
                        [2] <MstrMcName>
                        [3] <MstrMcType>
                        [4] <MstrTxDr>
                        [5] <sMstrTxDmn>
                        [6] <MstrUsr>
                        [7] <MstrRsrce>

 Returns: boolean value in the set of {0,1} where
          0 false
          1 true
 -------------------------------------------------------------------------

=cut

sub CheckDBBLOnResourceNode {
  #----------------
  # Get Arguments..
  #----------------
  my ( $rlrlsTuxResources ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];
  
  &logTrace ( 20236, "Subroutine <%s> called with:
  rlrlsTuxResources [%s]", $sMyName, $rlrlsTuxResources );

  #-----------------------------------------
  # Check if we have been passed some data..
  #-----------------------------------------
  unless ( scalar @$rlrlsTuxResources ) {
    &logError ( 20237,
    "<%s> No data to process. Cannot proceed.", $sMyName );
    return 0; # Failure..
  }

  #------------------- 
  # Define variables..
  #------------------- 
  my $iReturnCode = 1; # Setup return code to hopeful success..
  my $iFound = 0;      # Flag if DBBL is found..
  my $iCnt = 0;        # Count..

  my ( $sSysName, $sEnvFile, $sCommand, $sCommandArgs,
       $sUser, $iTimeout, $rlsTuxData, @lsCmdout ) = ();

  #------------------------------------
  # Process each candidate one by one..
  #------------------------------------
  foreach $rlsTuxData ( @$rlrlsTuxResources ) {
    &logTrace ( 20238,
    "<%s> Received data for Candidate No. [%s]
    SysName     [%s]
    MstrEnvFile [%s]
    MstrMcName  [%s]
    MstrMcType  [%s]
    MstrTxDr    [%s]
    MstrTxDmn   [%s]
    MstrUsr     [%s]
    MstrRsrce   [%s]", $sMyName, $iCnt + 1,
    $rlsTuxData->[0], $rlsTuxData->[1], $rlsTuxData->[2],
    $rlsTuxData->[3], $rlsTuxData->[4], $rlsTuxData->[5],
    $rlsTuxData->[6], $rlsTuxData->[7] );
    
    #------------------------------------
    # Run tmadmin on the requested node..
    #------------------------------------
    $sEnvFile = $rlsTuxData->[1];
    $sCommand = "$rlsTuxData->[4]/bin/tmadmin";
    $sSysName = $rlsTuxData->[0];
#---------------- Single definition -------------
    $sCommandArgs = "-r << !
psr -m $rlsTuxData->[2]
quit
!";
#---------------- Single definition -------------
    $sUser = $rlsTuxData->[6];
    $iTimeout = 10; # Reasonable..

    &logTrace ( 20239,
    "<%s> Preparing to run Command [%s] with Args [%s] with timeout [%s] secs",
    $sMyName, $sCommand, $sCommandArgs, $iTimeout );

    #----------------------------------
    # Now fire command on remote node..
    #----------------------------------
    $iReturnCode = Tuxedo::ExecRemoteTuxCmd ( $sEnvFile,
                                              $sUser,
                                              $sCommand,
                                              $sCommandArgs,
                                              \@lsCmdout,
                                              $iTimeout,
                                              $sSysName );

    $iReturnCode || return $iReturnCode; # Return, if failed..

    #-----------------------
    # Check if we got DBBL..
    #-----------------------
    if ( 1 == ( $iFound = grep { /^\bDBBL\b/o } @lsCmdout ) ) {
      &logTrace ( 20240,
      "<%s> Found DBBL process. Resource [%s] is the active TuxMaster",
      $sMyName, $rlsTuxData->[7] ); 

      #--------------------
      # Update ref. array..
      #--------------------
      @$rlrlsTuxResources = splice ( @$rlrlsTuxResources, $iCnt, 1 ); # Update..

      $iReturnCode = 1; # Set to true..
      last; # Get out..
    }
    else {
      &logTrace ( 20241,
      "<%s> Found [%s] instances of DBBL on [%s] in resource [%s]",
      $sMyName, $iFound, $rlsTuxData->[0], $rlsTuxData->[7] );
      $iReturnCode = 0; # Reset to indicate error..
    }

    ++$iCnt; # Increment..
  }

  return $iReturnCode;
}

=pod

 -------------------------------------------------------------------------
 Function : bool MigrateTuxedoMaster ( IN String TuxDir,
                                       IN String EnvFile,
                                       IN String User,
                                       IN String MachineName,
                                       IN Integer Timeout );

 Description : Performs migration of active master to this tux machine
 
 Parameters :

 sTuxDir         Absolute path to location of Tuxedo instance
 sEnvFile        Absolute path to enviroment file for Tuxedo
 sUser           UNIX user for Tuxedo instance
 sMachineName    Machine name registered with Tuxedo
 iTimeout        Timeout in secs.

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub MigrateTuxedoMaster {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sTuxDir,
       $sEnvFile,
       $sUser,
       $sMachineName,
       $iTimeout ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20242, "Subroutine <%s> called with:
  TuxDir      [%s]
  EnvFile     [%s] 
  User        [%s] 
  MachineName [%s]
  Timeout     [%s]",
  $sMyName,
  $sTuxDir, $sEnvFile, $sUser, $sMachineName, $iTimeout );

  #----------------------------------------
  # Set up return code to hopeful success..
  #----------------------------------------
  my $iReturnCode = 1;

  #------------------------------------
  # Generate args for running command..
  #------------------------------------
  my $sCommand = "$sTuxDir/bin/tmadmin";
  #---------------- Single definition -------------
  my $sCommandArgs = "-r << !
master -yes
psr -m $sMachineName
quit
!";
  #---------------- Single definition -------------
  my ( $iExitCode, @lsCmdout ) = ();

  &logWarn ( 20243, "<%s>
  MIGRATING MASTER with Command [%s] Args [%s]",
  $sMyName, $sCommand, $sCommandArgs );

  #--------------------------------
  # Attempt to migrate the master..
  #--------------------------------
  if ( Tuxedo::ExecLocalCmd ( $sUser,
                              $sEnvFile,
                              $sCommand,
                              $sCommandArgs,
                              \$iExitCode,
                              $iTimeout,
                              \@lsCmdout ) ) {
    #---------------------------------------------------------
    # Verify that we are currently hosting the active master..
    #---------------------------------------------------------
    if ( 1 == grep { /^\bDBBL\b/o } @lsCmdout ) {
      &logWarn ( 20244,
      "<%s> MIGRATION COMPLETED SUCCESSFULLY. Machine [%s] now hosts the DBBL",
      $sMyName, $sMachineName );
    }
    else {
      &logError ( 20245, "<%s> MIGRATION attempt FAILED!!",
      $sMyName );
      $iReturnCode = 0; # Failure..
    }
  }
  else {
    &logError ( 20246,
    "<%s> Could NOT perform MIGRATION procedure successfully.",
    $sMyName );
    $iReturnCode = 0; # Failure..
  }
  
  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool MigrateDBBLToBkp ( IN String MachineType,
                                    IN String TuxDomain,
                                    IN Integer Timeout,
                                    OUT StringRefernce TuxBkupAttr );

 Description  : Locates the resoure acting as the current backup
                machine in the Tuxedo domain, and migrates DBBL to
                it.

 We attempt to find  details  of  the  resource  within  the  same
 TuxDomain  that  is  currently  the  Backup, and migrate the DBBL
 to it.

 Parameters :

 $sMachineType    Bulletin board architecture
 $sTuxDomain      Name of the Tuxedo domain for this instance
 $iTimeeout       Time limit to attempt the migration
 $rlsTuxBkupAttr  Holds the following attributes of the new
                  active master, in this order:

                  BkupSysName, BkupEnvFile, BkupMcName,
                  BkupMcType, BkupTxDr, BkupUsr, BkupRsrce

 Returns : boolean value in the set of {0,1} where

           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub MigrateDBBLToBkp {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sMachineType,
       $sTuxDomain,
       $iTimeout,
       $rlsTuxBkupAttr ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20247, "Subroutine <%s> called with:
  MachineType     [%s],
  TuxDomain       [%s],
  Timeout         [%s]
  rlsTuxBkupAttr  [%s]",
  $sMyName, $sMachineType, $sTuxDomain, $iTimeout, $rlsTuxBkupAttr );

  #------------------------- 
  # Define local variables..
  #------------------------- 
  my $iReturnCode = 1; # Setup return code to hopeful success..
  my $iFound = 0;      # Flag if Backup is found..

  my ( $sBkupEnvFile,    $sBkupMcName,    $sBkupMcType,     $sBkupRsrce,
       $sBkupTxDr, $sBkupTxDmn, $sBkupUsr, $sBkupSysName, $i, $sCommand,
       $sCommandArgs, $iExitCode,  @lsHaresOut, @lsTemp, @lsCmdout ) = ();

  @$rlsTuxBkupAttr = (); # Truncate..

  &logTrace ( 20248,
  "<%s> Attempting an hares(1m) search for Backup with MachineType [%s]",
  $sMyName, $sMachineType );

  $sCommand = $sCmdHares; # Get global..
  $sCommandArgs = "-display -type Tuxedo -attribute " .
  "State EnvFile MachineName MachineType TuxDir TuxDomain User";

  #-----------------------------------------
  # Fire the command and collect the output..
  #-----------------------------------------
  @lsHaresOut =
  grep {
    ( s/\n// || 1 )                                   && # Chomp, and always make true..
    ( s/^(\S+).*\s(\S+)\s+ONLINE$/$1|$2/ && ++ $i )   || # If online, get composite, set flag..
    ( $i                                              && # If flag..
    /EnvFile|Machine(Name|Type)|TuxD(ir|omain)/o &&      # and it contains one of these..
    s/.*\s+// )                                       || # get the last field..
    ( $i && /User/                                    && # If flag, and has User..
    s/.*\s+/sprintf ("", -- $i )/e )                     # Get it, reset flag.
  } readpipe ( "$sCommand $sCommandArgs" );

  #------------------
  # Check exit code..
  #------------------
  if ( $iExitCode = $? >> 8 ) {
    &logError ( 20249,
    "<%s> Cmd [%s] with args [%s] failed with exit code [%s]",
    $sMyName, $sCommand, $sCommandArgs, $iExitCode );

    @lsHaresOut = (); # Truncate..
    return 0;         # Failure..
  }

  $i = 0;                               # Reset and reuse..
  $iFound = 0;                          # Reset.. ensure..

  if ( scalar @lsHaresOut ) {
    while ( scalar @lsHaresOut ) {
      @lsTemp = splice ( @lsHaresOut, 0, 7 ); # Extract in batches of 7..

      #---------------------------------------------
      # Get ResourceName, Node name from composite..
      #---------------------------------------------
      ( $lsTemp [7], $lsTemp [0] ) = $lsTemp [0] =~ /([^|]+)/g;
      $lsTemp [0] =~ s/.*://; # Remove GCO component from nodename..


      #-------------------------------------------------
      # Remove cluster component in name, if under GCO..
      #-------------------------------------------------
      $lsTemp [0] =~ s/.*://;

      ( $sBkupSysName,
        $sBkupEnvFile,
        $sBkupMcName,
        $sBkupMcType,
        $sBkupTxDr,
        $sBkupTxDmn,
        $sBkupUsr,
        $sBkupRsrce ) = @lsTemp;

      &logTrace ( 20250,
      "<%s> VCS query report from [%s] with args [%s]:
      ------------------------------------
      Res. No. [%s]   Resource Name [%s]
                     EnvFile       [%s] 
                     MachineName   [%s] 
                     MachineType   [%s] 
                     TuxDir        [%s] 
                     TuxDomain     [%s] 
                     User          [%s]  online on
                     System        [%s]",
      $sMyName,
      $sCommand, $sCommandArgs, ++ $i,
      $sBkupRsrce, $sBkupEnvFile, $sBkupMcName, $sBkupMcType,
      $sBkupTxDr, $sBkupTxDmn, $sBkupUsr, $sBkupSysName );

      #--------------------------------
      # Check if it is a likely match..
      #--------------------------------
      if ( $sBkupMcType eq $sMachineType && # Is the reqd. MachineType..
           $sBkupTxDmn  eq $sTuxDomain ) {  # From the same domain..

        &logTrace ( 20251,
        "<%s> The current backup Machine [%s] has been located",
        $sMyName, $sBkupMcName );

        #------------------------------
        # Attempt to migrate the DBBL..
        #------------------------------
        $sCommand = "$sBkupTxDr/bin/tmadmin";

        #---------------- Single definition -------------
        $sCommandArgs = "-r << !
master -yes
psr -m $sBkupMcName
quit
!";
        #---------------- Single definition -------------

        &logWarn ( 20252,
        "<%s> Attempting to migrate DBBL with Command [%s] Args [%s] to Machine [%s]",
        $sMyName, $sCommand, $sCommandArgs, $sBkupMcName );

        #----------------------------------
        # Now fire command on remote node..
        #----------------------------------
        if ( $iReturnCode = Tuxedo::ExecRemoteTuxCmd ( $sBkupEnvFile,
                                                       $sBkupUsr,
                                                       $sCommand,
                                                       $sCommandArgs,
                                                       \@lsCmdout,
                                                       $iTimeout,
                                                       $sBkupSysName ) ) {

          #---------------------------------------------------------
          # Verify that we are currently hosting the active master..
          #---------------------------------------------------------
          if ( 1 == grep { /^\bDBBL has been migrated\b/o } @lsCmdout ) {
            &logWarn ( 20253,
            "<%s> MIGRATION COMPLETED SUCCESSFULLY. Machine [%s] now hosts the DBBL",
            $sMyName, $sBkupMcName );

            @$rlsTuxBkupAttr = @lsTemp; # Update..
            $iReturnCode = 1;           # Success..
            last;                       # Get out..
          }
          else {
            &logError ( 20254, "<%s> MIGRATION attempt FAILED!! Cmd out [%s]",
            $sMyName, join ( chr 012, @lsCmdout ) );
            $iReturnCode = 0; # Failed..
          }
        }
      }
      else {
        $iReturnCode = 0; # Failed..
      }
    }
  }
  else { # Error!
    &logTrace ( 20255,
    "<%s> Found no ONLINE resources of type Tuxedo", $sMyName );
    $iReturnCode = 0; # Failure..
  }

  return $iReturnCode;
}


=pod

 --------------------------------------------------------------
 Function Name: bool GetDomainMembershipStatus ( IN String User,
                                                 IN String EnvFile,
                                                 IN String TuxDomain,
                                                 IN String MachineName,
                                                 IN String MachineType,
                                                 IN String ResourceName,
                                                 IN Integer Called,
                                                 IN Integer Timeout );

 Description  : Verifies if current machine is part of Tux
                domain. Attempts migration, if no active master present.
 
 Parameters :

 $sUser           UNIX user for Tuxedo instance
 $sEnvFile        Absolute path to enviroment file for Tuxedo
 $sTuxDomain      Domain if MachineType is Master,
 $sMachineName    Machine name registered with Tuxedo NonMaster, or Backup
 $sMachineType    Master, NonMaster, Backup or Single
 $sResourceName   Name of this resource
 $iCalled         Invocation count
 $iTimeout        Timeout in sec.

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 --------------------------------------------------------------

=cut

sub GetDomainMembershipStatus {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sUser,
       $sEnvFile,
       $sTuxDir,
       $sTuxDomain,
       $sMachineName,
       $sMachineType,
       $sResourceName,
       $iCalled,
       $iTimeout ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20256, "Subroutine <%s> called with:
  User         [%s]
  EnvFile      [%s]
  TuxDir       [%s]
  TuxDomain    [%s]
  MachineName  [%s]
  MachineType  [%s]
  ResourceName [%s]
  Called       [%s]
  Timeout      [%s]",
  $sMyName,
  $sUser, $sEnvFile, $sTuxDir, $sTuxDomain, $sMachineName,
  $sMachineType, $sResourceName, $iCalled, $iTimeout );

  #---------------------
  # Set up return code..
  #---------------------
  my $iReturnCode = 1; # Hopeful success..

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sCommand, $sCommandArgs, $sMstrSysName, $sMstrRsrce,
       @lsMstrAttr ) = ();
  my $iFound = 0;               # Count of BBL processes..
  my $sNodeName = ( uname )[1]; # Grab second element..
  
  #-------------------------------
  # Search for the active master..
  #-------------------------------
  if ( $iReturnCode = Tuxedo::WhereIsMasterActive ( $sTuxDomain,
                                                    $sTuxDir,
                                                    \@lsMstrAttr ) ) {
    &logTrace( 20257,
    "<%s> Found TuxMaster in domain [%s] with:
    Resource Name [%s]
    EnvFile       [%s]
    MachineName   [%s]
    MachineType   [%s]
    TuxDir        [%s]
    TuxDomain     [%s]
    User          [%s] online on
    VCS Node      [%s]",
    $sMyName, $sTuxDomain,
    $lsMstrAttr [7], $lsMstrAttr [1], $lsMstrAttr [2], $lsMstrAttr [3],
    $lsMstrAttr [4], $lsMstrAttr [5], $lsMstrAttr [6], $lsMstrAttr [0] ); 
  }
  else {
    &logError ( 20258,
    "<%s> Could not locate active master for domain [%s]",
    $sMyName, $sTuxDomain );
    $iReturnCode = 0; # Set to failure..

    #---------------------------------------
    # Attempt to migrate DBBL, if possible..
    #---------------------------------------
    if ( 1 == $iCalled &&                          # Just once..
         $sMachineType =~ /^(Master|Backup)$/o ) { # Acting backup..
      &logWarn ( 20259, "<%s> ATTEMPTING TO MIGRATE master", $sMyName );

      if ( $iReturnCode = Tuxedo::MigrateTuxedoMaster ( $sTuxDir,
                                                        $sEnvFile,
                                                        $sUser,
                                                        $sMachineName,
                                                        $iTimeout ) ) {
        &logInfo ( 20260,
        "<%s> Migration procedure successful. We are the new active master",
        $sMyName );

        #----------------------------------------------
        # Update history file for WhereIsMasterActive..
        #----------------------------------------------
        @lsMstrAttr = ( $sNodeName, $sEnvFile, $sMachineName,
                        $sMachineType, $sTuxDir, $sUser, $sResourceName );

        Tuxedo::UpdateTuxMasterHistory ( $sTuxDir,
                                         \@lsMstrAttr );
      }
      else {
        &logError ( 20261,
        "<%s> Migration FAILED! Quitting!", $sMyName );
      }
    }
    elsif ( 1 == $iCalled && scalar @lsMstrAttr ) { # NonMaster, then just once..
      #------------------------------------
      # Get details about the backup node..
      #------------------------------------
      while ( ( $sMstrSysName, $sMstrRsrce ) = splice ( @lsMstrAttr, 0, 2 ) ) {
        &logTrace ( 20262,
        "<%s> Probing resource [%s] on system [%s] to expedite master migration",
        $sMyName, $sMstrRsrce, $sMstrSysName );

        $sCommand = $sCmdHares; # Get global..
        $sCommandArgs = "-probe $sMstrRsrce -sys $sMstrSysName";
        system ( "$sCommand $sCommandArgs" ); # Probe..
      }
    }

    return $iReturnCode; # Return here..
  }

  #------------------------------------
  # Generate args for running command..
  #------------------------------------
  $sEnvFile = $lsMstrAttr [1];
  $sUser = $lsMstrAttr [6];
  $sCommand = "$lsMstrAttr[4]/bin/tmadmin";
#---------------- Single definition -------------
  $sCommandArgs = "-r << !
psr -m $sMachineName
quit
!";
#---------------- Single definition -------------
  my ( @lsCmdout, %hiBBLCount ) = ();
  my $sSysName = $lsMstrAttr [0]; # Where active master is..

  #----------------------------------
  # Now fire command on remote node..
  #----------------------------------
  $iReturnCode = Tuxedo::ExecRemoteTuxCmd ( $sEnvFile,
                                            $sUser,
                                            $sCommand,
                                            $sCommandArgs,
                                            \@lsCmdout,
                                            $iTimeout,
                                            $sSysName );

  $iReturnCode || return $iReturnCode; # Return, if failed..

  #--------------------------------------------------------------
  # Check if we found at least one BBL.. that's not PARTITIONED..
  #--------------------------------------------------------------
  grep {
          ( s/\n// || 1 )                 &&  # Chomp and make true..
          ! /^Prog Name|\bPARTITIONED\b/o &&  # Skip headers and partitioned..
          /\b(^BBL)\b/o                   &&  # Pick up all BBLs..
          ++ $hiBBLCount { $1 }               # And count them..
  } @lsCmdout;                                # From the command output..

  if ( $hiBBLCount { BBL } ) {
    &logTrace ( 20263,
    "<%s> Found [%s] counts of requisite BBL process.",
    $sMyName, $hiBBLCount{'BBL'} );
  }
  else {
    &logWarn ( 20264, "<%s> Could not locate any BBL on TuxMaster [%s].
    Are we PARTITIONED??", $sMyName, $lsMstrAttr [2] );
    $iReturnCode = 0 ; # Failure..
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool UseBackupBootFlag ( IN String User,
                                     IN String EnvFile,
                                     IN String MachineName );

 Description : Determines if '-b' flag is needed to boot this Machine.

 Parameters  :

 $sUser            UNIX user for Tuxedo instance
 $sEnvFile         Absolute path to enviroment file for Tuxedo
 $sMachineName     Machine name registered with Tuxedo

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub UseBackupBootFlag {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sUser,
       $sEnvFile,
       $sMachineName ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20265, "Subroutine <%s> called with:
  User        [%s]
  EnvFile     [%s]
  MachineName [%s]",
  $sMyName, $sUser, $sEnvFile, $sMachineName );

  #-------------------------------------
  # Set return code to hopeful success..
  #-------------------------------------
  my $iReturnCode = 1;

  #-------------------------
  # Define local variables..
  #-------------------------
  my $i = 0;
  my ( $sMaster, $sBackup, $sTuxConfig, $iExitCode,
       @lsCmdout, @lsTuxConfigData ) = ();

  #-------------------------------------
  # Get TUXCONFIG from user enviorment..
  #-------------------------------------
  my $sCommand = "/bin/echo";
  my $sCommandArgs = "\$TUXCONFIG";
  my $iTimeout = 5; # Reasonable..

  #------------------------------
  # Get TUXCONFIG file from env..
  #------------------------------
  if ( Tuxedo::ExecLocalCmd ( $sUser,
                              $sEnvFile,
                              $sCommand,
                              $sCommandArgs,
                              \$iExitCode,
                              $iTimeout,
                              \@lsCmdout ) ) {
    #--------------------------------------------
    # Check if we can locate the tuxconfig file..
    #--------------------------------------------
    $sTuxConfig = shift @lsCmdout; # Get o/p..

    if ( ! -r $sTuxConfig ) {
      &logTrace ( 20266, "<%s> Could not locate TUXCONFIG [%s] file",
      $sMyName, $sTuxConfig );
      return 0; # Failure..
    }
  }
  else {
     return 0; # Failure..
  }

  #--------------------------------------------------------------------
  # On reading the binary file tuxconfig, the line following a 'TAGENT'
  # is the machine that last hosted the   DBBL.    The next one is the
  # backup.
  #--------------------------------------------------------------------

  if ( open ( FILDES, $sTuxConfig ) ) {
    binmode FILDES; # Binary read..
    @lsTuxConfigData = map { /([\040-\176\s]{4,})/g } <FILDES>;
 
    close ( FILDES ) ||
    &logError ( 20267, "<%s> Close failed on TUXCONFIG [%s] after read",
    $sMyName, $sTuxConfig );
  }
  else {
    &logError ( 20268, "<%s> Open failed on TUXCONFIG [%s] for read",
    $sMyName, $sTuxConfig );
    return 0; # Failure..
  }

  ++ $i until $i > $#lsTuxConfigData ||          # Upper limit..
              $lsTuxConfigData [$i] eq 'TAGENT'; # Located..
  $sMaster = $lsTuxConfigData [$i + 1];          # Migrated DBBL last..
  $sBackup = $lsTuxConfigData [$i + 2];          # The backup..

  if ( $sMaster ) {
    &logTrace ( 20269,
    "<%s> Machine [%s] migrated DBBL last", $sMyName, $sMaster );
    $iReturnCode = 0 if $sMaster eq $sMachineName; # Active master..
  }

  $sBackup &&
  &logTrace ( 20270,
  "<%s> Machine [%s] is the backup machine", $sMyName, $sBackup );

  return $iReturnCode;
}

=pod

 -------------------------------------------------------------------------
 Function : bool ManageTuxIPCs ( IN Sring EnvFile,
                                 IN String TuxDir,
                                 IN String User
                                 IN Integer Timeout,
                                 IN Bool ClearHistory,
                                 OUT ListReference Cmdout );

 Description : Gets IPC resources for a Tuxedo machine.

 Parameters :

 $sEnvFile          Absolute path to environment file to source
 $sTuxDir           Absolute path to Tuxedo base installation directory
 $sUser             The dedicated UNIX login for this Tuxedo instance
 $iTimeout          Time limit to run UNIX command
 $bClearHistory     Behaviour history flag
 $rlsCmdout         Output of tmipcrm command

 ClearHistory = 0:

 After successful execution of tmipcrm, creates a history file with the
 command output. $rlsCmdOut contains the command output.

 ClearHistory = 1:

 If tmipcrm is executed successfully, $rlsCmdout contains the command
 output. If missing, $rlsCmdout contains history file contents. History
 file is removed if it exists.

 Returns : integer value in the set of {-1,0,1} where

           -1 false ( Critical error )

           1. Time-out error
           2. tmipcrm not executable

           0 false ( Ignorable error )

           ClearHistory = 0: Could not create history file.
           ClearHistory = 1: Could not get IPC resources.

           1 true ( Success )

           ClearHistory = 0: History file created successfully.
           ClearHistory = 1: IPC resources retrived successfully.

 -------------------------------------------------------------------------

=cut

sub ManageTuxIPCs {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sEnvFile,
       $sTuxDir,
       $sUser,
       $iTimeout,
       $bClearHistory,
       $rlsCmdout ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20271, "Subroutine <%s> called with:
  EnvFile        [%s]
  TuxDir         [%s]
  User           [%s]
  Timeout        [%s]
  ClearHistory   [%s]
  rlsCmdout      [%s]",
  $sMyName, $sEnvFile, $sTuxDir, $sUser, $iTimeout,
  $bClearHistory, $rlsCmdout );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;


  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sCommand, $sCommandArgs, $sLine, $sResourceType,
       $sResourceName, $sIPCDir, $sIPCFile, $sEntryPointName,
       $bUseHis, $iExitCode ) = ();

  #----------------------------------------------
  # Derive our resource type and resource names..
  #----------------------------------------------
  $sResourceType = ( $0 =~ m@([^/]+)@g ) [-2];
  $sResourceName = $ENV { VCS_LOG_RESOURCE_NAME };
  ( $sEntryPointName = $0 ) =~ s#.*/##; # Basename..

  $sIPCDir = "/tmp/.VRTS$sResourceType/$sResourceName/IPC";
  $sIPCFile = "$sIPCDir/tmipcrm.out";
  @$rlsCmdout = (); # Truncate..

  #------------------------------
  # Create PID folder if needed..
  #------------------------------
  if ( -d $sIPCDir ) {
    &logTrace ( 20272, "<%s> IPC folder [%s] exists",
    $sMyName, $sIPCDir );
  }
  elsif ( ! mkpath ( $sIPCDir, 0, 0700 ) ) {
    &logWarn ( 20273, "<%s> Failed to create IPC folder [%s]",
    $sMyName, $sIPCDir );
    return -1; # Cannot proceed..
  }

  #-----------------------
  # Initialise variables..
  #-----------------------
  $sCommand = "$sTuxDir/bin/tmipcrm";
  $sCommandArgs = "-n";

  #----------------------------------
  # Command available for execution..
  #----------------------------------
  if ( -x $sCommand ) {
    &logTrace ( 20274,
    "<%s> Getting IPCs for Tuxedo instance with Cmd [%s] Args [%s]",
    $sMyName, $sCommand, $sCommandArgs );
  
    if ( $iTimeout <= 5 ) { # Hard code..
      &logError ( 20275,
      "<%s> Insufficient time to clean up." ,
      $sMyName, abs ( $iTimeout ) );
  
      return -1; # Critical..
    }
    elsif ( $iReturnCode =
      Tuxedo::ExecLocalCmd ( $sUser,
                             $sEnvFile,
                             $sCommand,
                             $sCommandArgs,
                             \$iExitCode,
                             $iTimeout,
                             $rlsCmdout ) ) {
      &logTrace ( 20276,
      "<%s> Cmd [%s] with args [%s] executed successfully",
      $sMyName, $sCommand, $sCommandArgs );
    }
    elsif ( ! $iExitCode ) {
      &logError ( 20277, "<%s> Cmd [%s] with args [%s] timed out",
      $sMyName, $sCommand, $sCommandArgs );

      $iReturnCode = -1; # Critical..
      ++ $bUseHis;       # Set flag..
    }
    else {
      &logWarn ( 20278,
      "<%s> Cmd [%s] with args [%s] failed with exit code [%s].
      Cmd output [%s]",
      $sMyName, $sCommand, $sCommandArgs, $iExitCode,
      join ( chr 012, @$rlsCmdout ) );
  
      ++ $bUseHis; # Set flag..
      $iReturnCode = 0; # Failure..
    }
  }
  else { # tmipcrm(1m) missing..
    &logError ( 20279, "<%s> Tuxedo cmd [%s] is not executable", $sMyName, $sCommand );
    $iReturnCode = -1; # Critical..

    return $iReturnCode if ( $sEntryPointName eq 'online' ); # Critical..
    ++ $bUseHis; # Set flag..
  }

  #------------------
  # Check behaviour..
  #------------------
  if ( $bClearHistory ) {
    if ( $bUseHis ) {
      &logTrace ( 20280,
      "<%s> Using IPCHis file [%s] to retrieve IPCs for Tuxedo instance",
      $sMyName, $sIPCFile );

      if ( open ( IPCFILE, $sIPCFile ) ) {
        @$rlsCmdout = grep { s/\n// || 1 } <IPCFILE>; # Slurp..

        close ( IPCFILE ) ||
        &logError ( 20281,
        "<%s> Failed to close IPCHis file [%s] after read [%s]",
        $sMyName, $sIPCFile, $! );
      }
      else {
        &logTrace ( 20282,
        "<%s> Could not open IPCHis file [%s] for retrieving IPC resources [%s]",
        $sMyName, $sIPCFile, $! );

        $iReturnCode = 0; # Error..
      }
    }

    #-----------------
    # Unlink IPCFile..
    #-----------------
    if ( -e $sIPCFile ) {
      unlink ( $sIPCFile ) ||
      &logError ( 20283, "<%s> Failed to unlink IPCFile [%s]",
      $sMyName, $sIPCFile );
    }
  }
  elsif ( $iReturnCode ) { # Create history, if successful..
    #---------------------
    # Create IPCHis file..
    #---------------------
    &logTrace ( 20284,
    "<%s> Creating IPCHis file [%s]", $sMyName, $sIPCFile );
  
    if ( sysopen ( IPCFILE, $sIPCFile, O_WRONLY|O_SYNC|O_CREAT|O_TRUNC,
                   0600 ) ) {
      print IPCFILE join ( chr 012, @$rlsCmdout ), "\n"; # Update..
  
      unless ( close ( IPCFILE ) ) {
        &logWarn ( 20285,
        "<%s> Failed to close IPCHis file [%s] after update: [%s]",
        $sMyName, $sIPCFile, $! );
        $iReturnCode = 0;
      }
    }
    else {
      &logWarn ( 20286,
      "<%s> Failed to create IPCHis file [%s]: [%s]",
      $sMyName, $sIPCFile, $! );
      $iReturnCode = 0;
    }
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool RemoveIPCs ( IN String EnvFile
                              IN String TuxDir,
                              IN String User
                              IN Integer Timeout )

 Description : Clears IPC resources for a Tuxedo machine

 Parameters :

 $sEnvFile      Absolute path to environment file to source
 $sTuxDir       Absolute path to Tuxedo base installation directory
 $sUser         The dedicated UNIX login for this Tuxedo instance
 $iTimeout      Time limit to run UNIX command

 Returns : boolean value in the set of {0,1} where

           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub RemoveIPCs {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sEnvFile,
       $sTuxDir,
       $sUser,
       $iTimeout ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20287, "Subroutine <%s> called with:
  EnvFile     [%s]
  TuxDir      [%s]
  User        [%s]
  Timeout     [%s]",
  $sMyName, $sEnvFile, $sTuxDir, $sUser, $iTimeout );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #-------------------------
  # Define local variables..
  #-------------------------
  my $bClearHistory = 1;

  my ( $bFlag, $iId, $iExitCode, $sLine, $sEntryPointName,
       @lsCmdout, @liQids, @liShmIds, @liSemIds ) = ();

  ( $sEntryPointName = $0 ) =~ s#.*/##; # Basename..

  if ( $bSharedLogin ) {  # Check Global flag..
    if ( $iTimeout <= 5 ) { # Hard code..
      &logError ( 20288,
      "<%s> Insufficient time to clean up." ,
      $sMyName, abs ( $iTimeout ) );
  
      return -1; # Failure..
    }
    elsif ( 1 == ( $iExitCode =
      Tuxedo::ManageTuxIPCs ( $sEnvFile,
                              $sTuxDir,
                              $sUser,
                              $iTimeout,
                              $bClearHistory,
                              \@lsCmdout ) ) ) {
      &logTrace ( 20289,
      "<%s> Located valid Tuxedo IPC resources for removal", $sMyName );
    }
    elsif ( -1 == $iExitCode &&
            $sEntryPointName eq 'online' ) {
      return $iExitCode; # Critical..
    }
    elsif ( $lsCmdout [-1] &&
            $lsCmdout [-1] =~ /^CMDTUX_CAT:1504/ ) { # BBL is down..
      &logTrace ( 20290,
      "<%s> No valid Tuxedo IPC resources found for removal", $sMyName );
  
      $iReturnCode = 1;    # Re-set..
      return $iReturnCode; # Get out..
    }
  
    #-----------------------
    # Now parse the output..
    #-----------------------
    foreach $sLine ( @lsCmdout ) {
      $sLine =~ s/^\s*//g; # Strip leading spaces.
      $sLine =~ s/\s*$//;  # And the trailing ones..
  
      #-----------------
      # Message queues..
      #-----------------
      if ( $sLine =~ /^Message Queues:/ ... $sLine !~ /\S/ ) {
        push ( @liQids, $sLine ) if ( $sLine =~ /^\d+/ );
       ++ $bFlag;
      }
  
      #----------------
      # Shared memory..
      #----------------
      if ( $sLine =~ /^Shared Memory:/ ... $sLine !~ /\S/ ) {
        push ( @liShmIds, $sLine ) if ( $sLine =~ /^\d+/ );
        ++ $bFlag;
      }
  
      #-------------
      # Semaphores..
      #-------------
      if ( $sLine =~ /^Semaphores:/ ... $sLine !~ /\S/ ) {
        push ( @liSemIds, $sLine ) if ( $sLine =~ /^\d+/ );
        ++ $bFlag;
      }
    }
  
    #------
    # Log..
    #------
    if ( $bFlag ) {
      &logTrace ( 20291, "   <%s> Located following IPC resources:
  
      ----------------------
      Message Queue Ids [%s]
      Total [%s]

      Shared Memory Ids [%s]
      Total [%s]

      Semaphore Ids     [%s]
      Total [%s]
      ----------------------
      Total located     [%s]
      ----------------------",
      $sMyName,
      join ( ' ', @liQids ),   scalar @liQids,
      join ( ' ', @liShmIds ), scalar @liShmIds,
      join ( ' ', @liSemIds ), scalar @liSemIds,
      scalar @liQids + scalar @liShmIds + scalar @liSemIds );
    }
  
    #-----------------
    # Message Queues..
    #-----------------
    if ( scalar @liQids ) {
      &logTrace ( 20292, "<%s> Removing Msg Queues", $sMyName );
  
      if ( Proc::RemoveIPCResourcesForUser ( $sUser,
                                             O_IPCMSQ,
                                             \@liQids ) ) {
        &logTrace ( 20293,
        "<%s> Removed Msg Queue Ids [%s] Total [%s] successfully",
        $sMyName, join ( ' ', @liQids ), scalar @liQids );
      }
      else {
        &logError ( 20294,
        "<%s> Failed to remove all Msg Queues", $sMyName );
        $iReturnCode = 0; # Failure..
      }
    }
  
    #----------------
    # Shared memory..
    #----------------
    if ( scalar @liShmIds ) {
      &logTrace ( 20295, "<%s> Removing Shared Memory Ids", $sMyName );
  
      if ( Proc::RemoveIPCResourcesForUser ( $sUser,
                                             O_IPCSHM,
                                             \@liShmIds ) ) {
        &logTrace ( 20296,
        "<%s> Removed Shared Memory Ids [%s] Total [%s] successfully",
        $sMyName, join ( ' ', @liShmIds ), scalar @liShmIds );
      }
      else {
        &logError ( 20297,
        "<%s> Failed to remove all Shared Memories", $sMyName );
        $iReturnCode = 0; # Failure..
      }
    }
  
    #-------------
    # Semaphores..
    #-------------
    if ( scalar @liSemIds ) {
      &logTrace ( 20298, "<%s> Removing Semapores", $sMyName );
  
      if ( Proc::RemoveIPCResourcesForUser ( $sUser,
                                             O_IPCSEM,
                                             \@liSemIds ) ) {
        &logTrace ( 20299,
        "<%s> Removed Semaphore Ids [%s] Total [%s] successfully",
        $sMyName, join ( ' ', @liSemIds ), scalar @liSemIds );
      }
      else {
        &logError ( 20300,
        "<%s> Failed to remove all Semaphores", $sMyName );
        $iReturnCode = 0; # Failure..
      }
    }
  }
  elsif ( Proc::RemoveIPCResourcesForUser ( $sUser, # UNIX login unique..
                                            O_IPCALL ) ) {
    &logTrace ( 20301,
    "<%s> Removed all IPC resources for this instance successfully",
    $sMyName );
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool CleanBBofMachine ( IN String EnvFile
                                    IN String TuxDir,
                                    IN String MachineName,
                                    IN String User
                                    OUT ListReference CmdOut,
                                    IN Integer Timeout,
                                    IN String SysName );

 Description : Requests clean up bulletin board from active master.

 Parameters :

 $sEnvFile      Absolute path to enviroment file of active master
 $sTuxDir       Absolute path to location of Tuxedo instance of active master
 $sMachineName  Machine name registered with Tuxedo to be cleaned
 $sUser         UNIX user of active master
 $rlsCmdout     Holds output from remote command after successful execution
 $iTimeout      Time limit to run UNIX command
 $sSysName      VCS node name in cluster hosting the active master

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub CleanBBofMachine {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sEnvFile,
       $sTuxDir,
       $sMachineName,
       $sUser,
       $rlsCmdout,
       $iTimeout,
       $sSysName ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20302, "Subroutine <%s> called with:
  EnvFile     [%s]
  TuxDir      [%s]
  MachineName [%s]
  User        [%s]
  rlsCmdout   [%s]
  Timeout     [%s]
  NodeName    [%s]",
  $sMyName, $sEnvFile, $sTuxDir, $sMachineName,
  $sUser, $rlsCmdout, $iTimeout, $sSysName );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;
 
  #-----------------------
  # Initialise variables..
  #-----------------------
  my $sCommand = "$sTuxDir/bin/tmadmin";
  #---------------- Single definition -------------
  my $sCommandArgs = "<< !
pclean $sMachineName
quit
!";
  #---------------- Single definition -------------
  
  @$rlsCmdout = (); # Reset..

  #-------------------------------------
  # Now execute command on remote node..
  #-------------------------------------
  $iReturnCode = Tuxedo::ExecRemoteTuxCmd ( $sEnvFile,
                                            $sUser,
                                            $sCommand,
                                            $sCommandArgs,
                                            $rlsCmdout,
                                            $iTimeout,
                                            $sSysName );

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int GetEngineTimeout ( IN Bool UseMargin );

 Description : Calculates engine timeout for entry point
 
 Parameters  :

 $bUseMargin      Flag to indicate if safety margin is requested

                  boolean value in the set of {0,1} where
                  0 No safety margin needed
                  1 Safety margin to be used
 Returns:  Time in secs.
 -------------------------------------------------------------------------

=cut

sub GetEngineTimeout {
  #----------------
  # Get Arguments..
  #----------------
  my ( $bUseMargin ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20303, "Subroutine <%s> called with:
  UseMargin  [%s]",
  $sMyName, $bUseMargin );

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sAttribute, $sCommand, $sCommandArgs,
       $iEngineTimeout, $iExitCode ) = ();

  #----------------------------------
  # Grace margins from VCS timeouts..
  #----------------------------------
  my %hiGraceMargin = ( 'MonitorTimeout', 10,   # Seconds..
                        'OnlineTimeout',  10,
                        'OfflineTimeout', 10,
                        'CleanTimeout',   20 );

  #--------------------------------------------------
  # Attempt to get resource level overridden values..
  #--------------------------------------------------
  my $sResourceName = $ENV { VCS_LOG_RESOURCE_NAME };
  my ( $sEntryPointName, $sResourceType ) = reverse $0 =~ m@([^/]+)@g;
  ( $sAttribute = $sEntryPointName ) =~ s/(.*)/\u$1Timeout/;

  #-----------------------------------
  # Query if it has already been set..
  #-----------------------------------
  if ( defined $::giEPTimeout &&    # Defined..
       $::giEPTimeout =~ /^\d+/ ) { # to an integer..
    $iEngineTimeout = $::giEPTimeout;

    &logTrace ( 20324,
    "<%s> Using AEPTimeout [%s] as EP timeout value",
    $sMyName, $iEngineTimeout );
  }
  else {
    #-----------------------------------------
    # Form query string based on entry point..
    #-----------------------------------------
    $sCommand = $sCmdHares; # Get global..
    $sCommandArgs = "-value $sResourceName $sAttribute";
  
    #----------------------------------
    # Fire command and collect output..
    #----------------------------------
    ( $iEngineTimeout = readpipe ( "$sCommand $sCommandArgs 2>/dev/null" ) )
                      =~ s/\n//;
  
    #--------------------
    # Check exit status..
    #--------------------
    if ( $iExitCode = $? >> 8 ) {
      &logTrace ( 20304,
      "<%s> Command [%s] with Args [%s] failed with exitcode [%s].",
      $sMyName, $sCommand, $sCommandArgs, $iExitCode );
    }
    elsif ( ! $iEngineTimeout ) { # Undefined value..
      &logTrace ( 20305, "<%s> Could not get [%s] from engine",
      $sMyName, $sAttribute );
    }
    else { # Success..
      &logTrace ( 20306, "<%s> Got [%s] as [%s] sec from engine",
      $sMyName, $sAttribute, $iEngineTimeout );
    }
  
    #-----------------------------------------
    # Form query string based on entry point..
    #-----------------------------------------
    unless ( $iEngineTimeout ) {
      $sCommand = $sCmdHatype; # Get global..
      ( $sAttribute = $sEntryPointName ) =~ s/(.*)/\u$1Timeout/;
      $sResourceType = "Tuxedo" if $sResourceType !~ /\w/; # Ensure..
      $sCommandArgs = "-value $sResourceType $sAttribute";
    
      #----------------------------------
      # Fire command and collect output..
      #----------------------------------
      ( $iEngineTimeout = readpipe ( "$sCommand $sCommandArgs" ) ) =~ s/\n//;
    
      #--------------------
      # Check exit status..
      #--------------------
      if ( $iExitCode = $? >> 8 ) {
        &logTrace ( 20325,
        "<%s> Command [%s] with Args [%s] failed with exitcode [%s]",
        $sMyName, $sCommand, $sCommandArgs, $iExitCode );
        $iEngineTimeout = $bUseMargin ? $hiGraceMargin { $sAttribute } : 0;
      }
      elsif ( ! $iEngineTimeout ) { # Empty value..
        &logTrace ( 20326, "<%s> Could not get [%s] from engine",
        $sMyName, $sAttribute );
        $iEngineTimeout = $bUseMargin ? $hiGraceMargin { $sAttribute } : 0;
      }
      else { # Success..
        &logTrace ( 20327, "<%s> Got [%s] as [%s] sec from engine",
        $sMyName, $sAttribute, $iEngineTimeout );
      }
    }
    else {
      &logTrace ( 20328, "<%s> Got [%s] as [%s] sec from engine",
      $sMyName, $sAttribute, $iEngineTimeout );
    }
  }

  #----------------------------
  # Apply margin if requested.. 
  #----------------------------
  if ( $bUseMargin     &&                                    # Margin requested..
       $iEngineTimeout >= $hiGraceMargin { $sAttribute } ) { # And valid..
    $iEngineTimeout -= $hiGraceMargin { $sAttribute };       # Apply..

    &logTrace ( 20307,
    "<%s> Setting up [%s] sec margin. Returning [%s]",
    $sMyName, $hiGraceMargin { $sAttribute }, $iEngineTimeout );
  }

  return ( $iEngineTimeout ); # Secs..
}

1; # So PERL knows its included..
